/**
  * File: DomainMessagesHandler.java
  * Content: Class supporting the blocking behaviour casting messages on subgroups
  * Author: LuisM Pena
  * Date: 09th September 2001
  * Version: 0.51.00
  * Last change:
  *
  *
  **/

package sensei.domains;

import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.Message;
import sensei.middleware.domains.DomainMessage;
import sensei.middleware.domains.MemberStateException;
import sensei.middleware.domains.MemberStateExceptionReason;
import sensei.middleware.domains.SyncCastDomainMessageProcessedException;

import sensei.util.Debug;
import sensei.util.Error;

/**
 * Class supporting the blocking behaviour casting messages on subgroups
 */
public class DomainMessagesHandler
{

  //*************************************************************************************
  //**************************** CONSTRUCTOR ********************************************
  //*************************************************************************************

  public DomainMessagesHandler(int memberId, GroupHandler groupHandler, SubgroupsManager subgroupsManager)
  {
    Trace.code("DomainMessagesHandler.java -> DomainMessagesHandler ( int memberId , GroupHandler groupHandler , SubgroupsManager subgroupsManager )");
    this.memberId = memberId;
    this.groupHandler=groupHandler;
    this.subgroupsManager=subgroupsManager;
    messageSemaphore = new MessageSemaphore ();
    messageId=0;
  }

  //*************************************************************************************
  //**************************** FREE RESOURCES *****************************************
  //*************************************************************************************
  /**
   * Frees any resources, the object can not be used anymore
   */
  public synchronized void freeResources()
  {
    Trace.code("DomainMessagesHandler.java -> synchronized void freeResources ( )");
    if (messageSemaphore!=null)
    {
      messageSemaphore.freeResources();
      messageSemaphore=null;
      groupHandler=null;
      subgroupsManager=null;
    }
  }

  public void startTransaction(TransactionsManager manager)
  {
    Trace.code("DomainMessagesHandler.java -> void startTransaction ( TransactionsManager manager )");
    transactionsManager=manager;
  }

  public void endTransaction()
  {
    Trace.code("DomainMessagesHandler.java -> void endTransaction ( )");
    transactionsManager=null;
  }

  //*************************************************************************************
  //**************************** SEND MESSAGE *******************************************
  //*************************************************************************************

  public boolean sendMessage(int target, Message msg)
  {
    Trace.code("DomainMessagesHandler.java -> boolean sendMessage ( int target , Message msg )");
    boolean ret;
    try
    {
      ret = groupHandler.sendMessage(target, msg);
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
      ret = false;
    }
    return ret;
  }


  //*************************************************************************************
  //**************************** CAST MESSAGE *******************************************
  //*************************************************************************************

  public boolean castMessage(Message msg)
  {
    Trace.code("DomainMessagesHandler.java -> boolean castMessage ( Message msg )");
    if (msg instanceof DomainMessage)
      ((DomainMessage)msg).waitReception=false;
    return directCastMessage(transactionsManager, msg);
  }

  boolean directCastMessage(TransactionsManager manager, Message msg)
  {
    Trace.code("DomainMessagesHandler.java -> boolean directCastMessage ( TransactionsManager manager , Message msg )");
    boolean ret;
    try
    {
      if (manager==null)
        ret = groupHandler.castMessage(msg);
      else
        ret = manager.castMessage(msg);
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
      ret = false;
    }
    return ret;
  }

  //*************************************************************************************
  //**************************** SYNC CAST DOMAIN MESSAGE *******************************
  //*************************************************************************************

  public boolean syncCastDomainMessage(DomainMessage message, boolean normalProcessing) throws MemberStateException
  {
    Trace.code("DomainMessagesHandler.java -> boolean syncCastDomainMessage ( DomainMessage message , boolean normalProcessing )");
    int id = getMessageId();
    message.id = id;
    message.waitReception=true;

    Trace.syncCastDomainMessageStarts(id, normalProcessing);

    //if the message is transactionable, it must be sent through the transaction manager, if there is one
    TransactionsManager manager = message.unTransactionable && message.unqueuedOnST? null : transactionsManager;

    //create an entry on the semaphore
    MessageSemaphore.Entry semaphoreEntry = messageSemaphore.create(id, !normalProcessing);

    boolean sentThisView = directCastMessage(manager, message);

    //wait for the message to be received on the processCastMessage side
    if (!messageSemaphore.join(semaphoreEntry))
      throw new MemberStateException(MemberStateExceptionReason.MemberExcluded);

    //if it's not normal processing, store this entry, to be removed after a call to syncCastDomainMessageProcessed
    Debug.assert(currentEntry==null,Consts.AREA,"DomainMessagesHandler::syncCastDomainMessage");

    if (!normalProcessing)
      currentEntry=semaphoreEntry;

    Trace.syncCastDomainMessageEnded(id, normalProcessing);

    return sentThisView;
  }

  public void syncCastDomainMessageProcessed() throws MemberStateException, SyncCastDomainMessageProcessedException
  {
    Trace.code("DomainMessagesHandler.java -> void syncCastDomainMessageProcessed ( )");
    if (currentEntry==null)
      throw new SyncCastDomainMessageProcessedException();
    MessageSemaphore.Entry use = currentEntry;
    currentEntry=null;
    if (!messageSemaphore.remove(use))
      throw new MemberStateException(MemberStateExceptionReason.MemberExcluded);
    Trace.syncCastDomainMessageProcessed(use.id);
  }

  //*************************************************************************************
  //**************************** PROCESS CAST MESSAGE EVENT *****************************
  //*************************************************************************************

  public void processCastMessageEvent(int sender, Message message)
  {
    Trace.code("DomainMessagesHandler.java -> void processCastMessageEvent ( int sender , Message message )");
    try
    {
      if (message instanceof DomainMessage)
      {
        DomainMessage domainMessage = (DomainMessage) message;

        Trace.processingDomainMessage(domainMessage.id);

        int subgroup = domainMessage.subgroup;

        boolean process=true;
        MessageSemaphore.Entry entry = null;

        if (domainMessage.waitReception && (sender==memberId))
        {
          entry = messageSemaphore.release(domainMessage.id);
          process=entry!=null;
        }

        if (process)
        {
          if (subgroup==Consts.EVERY_SUBGROUP)
            subgroupsManager.castMessage(sender, message);
          else
            subgroupsManager.castMessage(subgroup, sender, message);
          if (entry!=null)
            messageSemaphore.remove(entry);
        }
        Trace.processedDomainMessage(domainMessage.id);
      }
      else
      {
        //no domain message
        subgroupsManager.castMessage(sender, message);
      }
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }
  }


  //*************************************************************************************
  //**************************** PROCESS PTP MESSAGE EVENT ******************************
  //*************************************************************************************

  public void processPTPMessageEvent(int sender, Message message)
  {
    Trace.code("DomainMessagesHandler.java -> void processPTPMessageEvent ( int sender , Message message )");
    try
    {
      if (message instanceof DomainMessage)
      {
        DomainMessage domainMessage = (DomainMessage) message;

        int subgroup = domainMessage.subgroup;

        if (subgroup == Consts.EVERY_SUBGROUP)
          subgroupsManager.sendMessage(sender, message);
        else
          subgroupsManager.sendMessage(subgroup, sender, message);

      }
      else
      {
        //no domain message
        subgroupsManager.sendMessage(sender, message);
      }
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }
  }

  //*************************************************************************************
  //**************************** GET MESSAGE ID *****************************************
  //*************************************************************************************

  /**
    * Returns a unique message id inside this groupHandler
    **/
  synchronized int getMessageId()
  {
    Trace.code("DomainMessagesHandler.java -> synchronized int getMessageId ( )");
    return ++messageId;
  }

  //*************************************************************************************
  //**************************** DATA MEMBERS *******************************************
  //*************************************************************************************

  //there is a GroupHandler behalf the DomainGroupHandler, null when it is not yet activated
  GroupHandler groupHandler;

  //a copy of this member identity is kept, as is used when casting subgroup messages
  int memberId;

  //semaphore on state reception/processing
  MessageSemaphore messageSemaphore;
  MessageSemaphore.Entry currentEntry;

  //used to assign message numbers
  int messageId;

  //The subgroups handler, that propagates the messages to the subgroups
  SubgroupsManager subgroupsManager;

  //In case of transactions, messages are sent through the transactionManager
  TransactionsManager transactionsManager;
}
