/**
  * File: EventsQueue.java
  * Content: Events coming from the GMS are queued through this class. It allows to block messages
  *   and the rest of requirements required to implement the state transfer
  * Author: LuisM Pena
  * Date: 16th April 2001
  * Version: 0.51.00
  * Last change:
  *
  *
  **/


package sensei.domains;

import sensei.middleware.gms.Message;
import sensei.middleware.gms.View;
import sensei.middleware.domains.DomainGroupHandler;
import sensei.middleware.domains.DomainMessage;

import sensei.util.Debug;
import sensei.util.Error;

import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;

/**
 * Class to queue every event coming from the GMS. It allows to block messages and implements
 * the rest of requirements required by the state transfer logic
 */
public class EventsQueue extends Thread
{

  //*************************************************************************************//
  //**************************** CONSTRUCTOR ********************************************//
  //*************************************************************************************//

  public EventsQueue(EventsQueueUser user)
  {
    Trace.code("EventsQueue.java -> EventsQueue ( EventsQueueUser user )");
    this.user=user;
    events = new ArrayList();
    Q1 = new ArrayList(); //messages are initially blocked!
    Q2 = new ArrayList();
    alarm=false;
    everyMessageSent=true;

    start();
  }

  //*************************************************************************************
  //**************************** FREE RESOURCES *****************************************
  //*************************************************************************************
  /**
   * Frees any resources, the object can not be used anymore
   */
  public void freeResources()
  {
    Trace.code("EventsQueue.java -> void freeResources ( )");
    //resources are automatically released after receiving the expulsion event!!
    Debug.assert(events==null, Consts.AREA, "EventsQueue::freeResources::1");
  }

  //*************************************************************************************//
  //**************************** UNBLOCK ************************************************//
  //*************************************************************************************//

  /**
   * Unblock the messages from previous views and/or from this view.
   * If thisView is false, the alarm is set, otherwise is reset
   */
  public synchronized void unblock(boolean previousViews, boolean thisView)
  {
    Trace.code("EventsQueue.java -> synchronized void unblock ( boolean previousViews , boolean thisView )");
    Debug.assert(events!=null, Consts.AREA, "EventsQueue::unblock::1"); //already expulsed
    Debug.assert(previousViews || thisView, Consts.AREA, "EventsQueue::unblock::2");
    Debug.assert(Q1!=null,Consts.AREA, "EventsQueue::unblock::3"); //messages should be blocked!
    //it cannot be unblocked a view if previous views are blocked
    Debug.assert(previousViews || Q2.isEmpty(),Consts.AREA, "EventsQueue::unblock::6");

    Trace.queueUnblock(previousViews, thisView);

    if (thisView)
      dump(Q1);

    if (previousViews)
    {
      dumpQ2();
      alarm=true;
    }
    if (thisView)
    {
      alarm=false;
      Q1=null;
    }

    Trace.queueUnblocked(alarm, Q1!=null);
  }

  //*************************************************************************************//
  //**************************** NEW MESSAGE ********************************************//
  //*************************************************************************************//

  public void newCastMessage(int sender, Message message)
  {
    Trace.code("EventsQueue.java -> void newCastMessage ( int sender , Message message )");
    newMessage(sender, message, false, false);
  }

  public void newPTPMessage(int sender, Message message)
  {
    Trace.code("EventsQueue.java -> void newPTPMessage ( int sender , Message message )");
    newMessage(sender, message, true, false);
  }

  public void newTransactionMessage(int sender, Message message)
  {
    Trace.code("EventsQueue.java -> void newTransactionMessage ( int sender , Message message )");
    newMessage(sender, message, false, true); //always cast!!
  }

  /**
   * Override blocks should be set only if the message should be processed despite any block due
   * to pending state transfers
   */
  synchronized void newMessage(int sender, Message message, boolean ptp, boolean transactionMessage)
  {
    Trace.code("EventsQueue.java -> synchronized void newMessage ( int sender , Message message , boolean ptp , boolean transactionMessage )");
    Debug.assert(events!=null, Consts.AREA, "EventsQueue::newMessage::1"); //already expulsed

    //if message is a transactionMessage, it must be always handled though the events queue, in order to be
    //sent on a different thread!

    if ((Q1==null) && !transactionMessage)//messages are non blocked
    {
      waitUntilMessagesAreSent();
      sendMessage(sender, message, ptp);
    }
    else
    {
      //if the view is blocked and the message is not of the type blocked, the message is added to the
      //queued list, otherwise it goes directly to the event list
      MessageEvent event = new MessageEvent(sender, message, ptp);

      if (transactionMessage || ((message instanceof DomainMessage) && (((DomainMessage)message).unqueuedOnST)))
      {
        events.add(event);
        everyMessageSent = false;
        notifyAll();
      }
      else
      {
        Q1.add(event);
      }
    }
    Trace.code("EventsQueue.java -> -- synchronized void newMessage ( int sender , Message message , boolean ptp )");
  }

  //*************************************************************************************//
  //**************************** NEW VIEWS **********************************************//
  //*************************************************************************************//

  public synchronized void expulsed()
  {
    Trace.code("EventsQueue.java -> synchronized void expulsed ( )");
    Debug.assert(events!=null, Consts.AREA, "EventsQueue::expulsed::1"); //already expulsed

    Trace.queueExpulsionEventArrived();

    waitUntilMessagesAreSent();

    user.processExpulsionEvent();

    events=null;

    notifyAll();
    Trace.code("EventsQueue.java -> -- synchronized void expulsed ( )");
  }

  public void changingView()
  {
    Trace.code("EventsQueue.java -> void changingView ( )");

    Debug.assert(events!=null, Consts.AREA, "EventsQueue::changingView::1"); //already expulsed

    Trace.queueChangingViewEventArrived();

    //deactivate first the alarm
    alarm=false;

    //wait til every message has been processed. Just to avoid that the user receives events and at the
    //same time the questions from the changingViewEvent
    waitUntilMessagesAreSent();

    //what to do with the currently blocked messages?
    toChangeView();

    //wait til every message that can be unlocked has been processed
    waitUntilMessagesAreSent();

    Trace.code("EventsQueue.java -> -- void changingView ( )");
  }

  public void newView(View view)
  {
    Trace.code("EventsQueue.java -> void newView ( View view )");

    Debug.assert(events!=null, Consts.AREA, "EventsQueue::newView::1"); //already expulsed

    Trace.queueNewViewEventArrived();

    //wait til every message has been processed. Just to avoid that the user receives events and at the
    //same time the questions from the newViewEvent
    waitUntilMessagesAreSent();

    //what to do with the currently blocked messages?
    preprocessView(view);

    //wait til every message that can be unlocked has been processed
    waitUntilMessagesAreSent();

    //and then install the view!
    user.processViewEvent(view);

    alarm = true; //always activated!
    synchronized(this){notifyAll();} //wake up the run thread, to send probably the alarm to the user

    Trace.code("EventsQueue.java -> -- void newView ( View view )");
  }

  /**
   * Asks to the user what reaction to have against the new view, with respect to the
   * blocking of messages.
   */
  void toChangeView()
  {
    Trace.code("EventsQueue.java -> void toChangeView ( )");
    int answer = user.changingViewEvent();
    switch (answer)
    {
      case EventsQueueUser.CONTINUE_BLOCKING_PREVIOUS:
        break;

      case EventsQueueUser.DISCARD_PREVIOUS_MESSAGES:
        Q1.clear();
        Q2.clear();
        break;

      case EventsQueueUser.UNBLOCK_ALL_PREVIOUS_VIEWS:
        Q2.add(Q1);
        dumpQ2();
        Q1=null;
        break;

      case EventsQueueUser.UNBLOCK_PREVIOUS_VIEW:
        Debug.assert(Q2.isEmpty(), Consts.AREA, "EventsQueue::toChangeView::1");
        //this answer can be given even if here is no messages blocked
        if (Q1!=null)
        {
          dump(Q1);
          Q1=null;
        }
        break;

      default:
        Error.error(Consts.AREA, "Received wrong answer from user application: EventsQueueUser::toChangeView");
        break;
    }
  }

  /**
   * Asks to the user what reaction to have against the new view, with respect to the
   * blocking of messages.
   */
  void preprocessView(View view)
  {
    Trace.code("EventsQueue.java -> void preprocessView ( View view )");
    int answer = user.newViewEvent(view);
    switch (answer)
    {
      case EventsQueueUser.CONTINUE_BLOCKING_PREVIOUS:
        Q2.add(Q1);
        Q1=null;
        break;

      case EventsQueueUser.DISCARD_PREVIOUS_MESSAGES:
        Q1.clear();
        Q2.clear();
        break;

      case EventsQueueUser.UNBLOCK_ALL_PREVIOUS_VIEWS:
        Q2.add(Q1);
        dumpQ2();
        Q1=null;
        break;

      case EventsQueueUser.UNBLOCK_PREVIOUS_VIEW:
        Debug.assert(Q2.isEmpty(), Consts.AREA, "EventsQueue::preprocessView::1");
        //this answer can be given even if here is no messages blocked
        if (Q1!=null)
        {
          dump(Q1);
          Q1=null;
        }
        break;

      default:
        Error.error(Consts.AREA, "Received wrong answer from user application: EventsQueueUser::preprocessView");
        break;
    }

    //create a new Q1 list if null
    if (Q1==null)
      Q1 = new ArrayList();

    Trace.queuePreprocessedView(answer);
  }

  //*************************************************************************************//
  //**************************** WAIT UNTIL MESSAGES ARE SENT ***************************//
  //*************************************************************************************//

  synchronized void waitUntilMessagesAreSent()
  {
    Trace.code("EventsQueue.java -> synchronized void waitUntilMessagesAreSent ( )");
    while(!everyMessageSent)
      try{wait();}catch(InterruptedException ex){}
    Trace.code("EventsQueue.java -> -- synchronized void waitUntilMessagesAreSent ( )");
  }

  //*************************************************************************************//
  //**************************** THREAD ABSTRACT METHODS ********************************//
  //*************************************************************************************//

  public void run()
  {
    Trace.code("EventsQueue.java -> void run ( )");

    while (true)
    {
      try
      {
        synchronized (this)
        {
          while ((events!=null) && (events.isEmpty()))
          {
            Trace.queueEmptyEventList(alarm);
            if (alarm)
            {
              user.processAlarmEvent();
              alarm=false;
            }
            else //perhaps user.processAlarmEvent has produced events!
              wait(); //after a view is installed, this wait() is waked up, but still without events->repeat loop&alarm!
          }
        }

        if (events==null) //expulsed
          break;
        processEvents();
      }
      catch(InterruptedException iex)
      {
        Error.unhandledException(Consts.AREA,iex);
      }
    }

    Trace.queueProcessStopping();

    synchronized(this)
    {
      Q2.clear();
      if (Q1!=null)
        Q1.clear();
      Q1=Q2=null;
      user=null;
    }
  }

  void processEvents()
  {
    Trace.code("EventsQueue.java -> void processEvents ( )");
    Trace.queueProcessEvents();

    while(true)
    {
      MessageEvent messageEvent=null;
      synchronized(this)
      {
        if (events.isEmpty())
        {
          everyMessageSent = true;
          notifyAll();
          break;
        }
        else
          messageEvent = (MessageEvent) (events.remove(0));
      }
      Trace.queueSendsMessage(messageEvent.ptp);
      sendMessage(messageEvent.sender, messageEvent.message,messageEvent.ptp);
    }

    Trace.queueProcessEventsEnded();
  }

  void sendMessage(int sender, Message message, boolean ptp)
  {
    Trace.code("EventsQueue.java -> void sendMessage ( int sender , Message message , boolean ptp )");
    if (ptp)
      user.processPTPMessageEvent(sender, message);
    else
      user.processCastMessageEvent(sender, message);
  }

  //*************************************************************************************//
  //**************************** DUMP METHODS *******************************************//
  //*************************************************************************************//

  /**
   * Dumps Q2 into the events list: it is dumped on inverse order, to process first the first views
   */
  synchronized void dumpQ2()
  {
    Trace.code("EventsQueue.java -> synchronized void dumpQ2 ( )");
    Trace.queueQ2Dumped(Q2.size());

    ListIterator iterator=Q2.listIterator(Q2.size());
    while(iterator.hasPrevious())
      dump((List) (iterator.previous()));
    Q2.clear();
  }

  /**
   * Dumps the given list into the events list: it is dumped on inverse order, to process first the first views
   */
  synchronized void dump(List list)
  {
    Trace.code("EventsQueue.java -> synchronized void dump ( List list )");
    int size = list.size();
    Trace.queueDumped(size);
    if (size>0)
    {
      ListIterator iterator=list.listIterator(list.size());
      while(iterator.hasPrevious())
        events.add(0, iterator.previous());
      list.clear();
      everyMessageSent=false;
      notifyAll();
    }
  }


  //*************************************************************************************//
  //**************************** INNER CLASSES ******************************************//
  //*************************************************************************************//

  class MessageEvent
  {
    Message message;
    boolean ptp;
    int sender;
    public MessageEvent(int sender, Message message, boolean ptp)
    {
      this.sender=sender;
      this.message=message;
      this.ptp=ptp;
    }
  }

  //*************************************************************************************//
  //**************************** DATA MEMBERS *******************************************//
  //*************************************************************************************//

  List events; //ordered list of events, can be unqueued messages or queues of messages
  List Q1; //ordered list of messages, containing the messages blocked on the current view. If it is null,
            //no blocking is performed
  List Q2; //ordered list of list of messages, containg the messages blocked on previous views
  boolean alarm, everyMessageSent;
  EventsQueueUser user;
}
