/**
  * File: GroupMonitorImpl.java
  * Content: Monitor implementation on a replicated environment
  * Author: LuisM Pena
  * Date: 03rd October 2000
  * Version: 0.51.00
  * Last change:
  *             6th November 2001: rewritten with state transfer support
  *             12th November 2001: state transfer support changes: monitor messages must
  *                   be received during state transfers, to fully support the notion of transactions
  *                   The state sent is the state when the view is installed
  *
  *
  **/

package sensei.domains;

import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.Message;
import sensei.middleware.gms.View;
import sensei.middleware.domains.DomainGroupHandler;
import sensei.middleware.domains.DynamicSubgroupInfo;
import sensei.middleware.domains.GroupMonitorBaseImpl;
import sensei.middleware.domains.MonitorException;
import sensei.middleware.domains.MemberStateException;
import sensei.middleware.domains.MemberStateExceptionReason;
import sensei.middleware.domains.State;
import sensei.middleware.domains.StateTransferType;
import sensei.middleware.domains.SubgroupsHandlerException;

import sensei.middleware.util.ObjectsHandling;

import sensei.util.Debug;

/**
  * Monitor implementation on a replicated environment
  **/
public class GroupMonitorImpl extends GroupMonitorBaseImpl
{

//*************************************************************************************//
//**************************** CONSTRUCTOR ********************************************//
//*************************************************************************************//

  /**
   * Creates a groupMonitor that must still receive the state.
   * The monitor can be created after the group has received an event subgroupCreated, or before
   * registering the object in the domain. In the last case, use register after this constructor
   */
  public GroupMonitorImpl(int subgroupId, DomainGroupHandler groupHandler) throws Exception
  {
    Trace.code("GroupMonitorImpl.java -> GroupMonitorImpl ( int subgroupId , DomainGroupHandler groupHandler )");
    this.subgroupId=subgroupId;
    this.groupHandler=groupHandler;
    monitor = new MonitorImpl(subgroupId, groupHandler);
  }

  /**
   * Creates a groupMonitor that registers itself dynamically on the specified groupHandler
   */
  public GroupMonitorImpl(DynamicSubgroupInfo dynamicSubgroupInfo, DomainGroupHandler groupHandler)
    throws SubgroupsHandlerException, MemberStateException, Exception
  {
    Trace.code("GroupMonitorImpl.java -> GroupMonitorImpl ( DynamicSubgroupInfo dynamicSubgroupInfo , DomainGroupHandler groupHandler )");
    this.groupHandler=groupHandler;
    monitor = new MonitorImpl(sensei.domains.Consts.EVERY_SUBGROUP, groupHandler);
    subgroupId = groupHandler.createSubgroup(dynamicSubgroupInfo, thisGroupMonitor);
    monitor.setSubgroupId(subgroupId);
    monitor.assumeState();
  }

  /**
   * Register the subgroup statically on the DomainsGroupHandler
   */
  public void register() throws MemberStateException, SubgroupsHandlerException, Exception
  {
    Trace.code("GroupMonitorImpl.java -> void register ( )");
    if (monitor==null)
      throw new MemberStateException(MemberStateExceptionReason.MemberExcluded);
    else
      groupHandler.registerSubgroup(subgroupId, thisGroupMonitor);
  }

  /**
   * Sends a request to create dynamically a GroupMonitor subgroup
   */
  public static void register(DynamicSubgroupInfo dynamicSubgroupInfo, DomainGroupHandler groupHandler)
    throws MemberStateException, SubgroupsHandlerException, Exception
  {
    Trace.code("GroupMonitorImpl.java -> void register ( DynamicSubgroupInfo dynamicSubgroupInfo , DomainGroupHandler groupHandler )");
    groupHandler.castSubgroupCreation(dynamicSubgroupInfo);
  }

  void freeResources()
  {
    Trace.code("GroupMonitorImpl.java -> void freeResources ( )");
    if (monitor!=null)
    {
      try{deactivate();}catch(Exception ex){}
      groupHandler=null;
      monitor.freeResources();
      monitor=null;
    }
  }

  //*************************************************************************************
  //**************************** LOCK OPERATIONS *****************************************
  //*************************************************************************************

  public void lock() throws MemberStateException
  {
    Trace.code("GroupMonitorImpl.java -> void lock ( )");
    monitor.lock();
  }

  /**
    * Unlocks the monitor
    **/
  public void unlock() throws MonitorException, MemberStateException
  {
    Trace.code("GroupMonitorImpl.java -> void unlock ( )");
    monitor.unlock();
  }

  //*************************************************************************************
  //**************************** STATE OPERATIONS ***************************************
  //*************************************************************************************

  public State getState()
  {
    Trace.code("GroupMonitorImpl.java -> State getState ( )");
    return monitor.getState();
  }

  public void setState(State s)
  {
    Trace.code("GroupMonitorImpl.java -> void setState ( State s )");
    monitor.setState(s);
  }

  public void assumeState()
  {
    Trace.code("GroupMonitorImpl.java -> void assumeState ( )");
    monitor.assumeState();
  }

  /***********************************************************************************************/
  /********************** METHODS ON THE APPLICATION GROUP MEMBER INTERFACE JAVA INTERFACE *******/
  /***********************************************************************************************/

  public void changingView(){}
  public void processPTPMessage(int sender, Message message){}

  public void memberAccepted(int identity, GroupHandler handler, View view)
  {
    Trace.code("GroupMonitorImpl.java -> void memberAccepted ( int identity , GroupHandler handler , View view )");
    Debug.assert(ObjectsHandling.areEquivalent(groupHandler, handler), Consts.AREA, "GroupMonitorImpl::memberAccepted::1");
    MonitorImpl impl=monitor;
    Debug.assert(impl!=null,Consts.AREA, "GroupMonitorImpl::memberAccepted::2");
    if (impl!=null)
      impl.memberAccepted(identity);
  }

  public void excludedFromGroup()
  {
    Trace.code("GroupMonitorImpl.java -> void excludedFromGroup ( )");
    monitor.excludedFromGroup();
    freeResources();
  }

  public void installView(View view)
  {
    Trace.code("GroupMonitorImpl.java -> void installView ( View view )");
    monitor.installView(view);
  }

  public void processCastMessage(int sender, Message message)
  {
    Trace.code("GroupMonitorImpl.java -> void processCastMessage ( int sender , Message message )");
    monitor.processCastMessage(sender, message);
  }

  //*************************************************************************************//
  //**************************** DATA MEMBERS *******************************************//
  //*************************************************************************************//

  MonitorImpl monitor;
  protected DomainGroupHandler groupHandler; //in case of a late registration
  protected int subgroupId; //in case of a late registration
}
