/**
  * File: MembersOnTransferExcludedFromGroupImpl.java
  * Content: Implementation middleware-independent of the DomainGroupHandler interface with behaviour
  *          of Members on Transfer Excluded From Group
  * Author: LuisM Pena
  * Date: 03rd October 2000
  * Version: 0.51.00
  * Last change:
  *
  *
  **/

package sensei.domains;

import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.Message;
import sensei.middleware.gms.View;
import sensei.middleware.domains.BehaviourOnViewChanges;
import sensei.middleware.domains.DomainGroupHandler;
import sensei.middleware.domains.EndOfTransactionMessage;
import sensei.middleware.domains.MemberStateException;
import sensei.middleware.domains.MessageC;
import sensei.middleware.domains.MessageE;
import sensei.middleware.domains.MessageP;
import sensei.middleware.domains.MessageS;
import sensei.middleware.domains.Monitor;
import sensei.middleware.domains.MonitorException;
import sensei.middleware.domains.TransactionException;

import sensei.util.Debug;

import java.util.HashMap;
import java.util.Map;

/**
 * Implementation middleware-independent of the DomainGroupHandler interface with behaviour
 * of Members on Transfer Excluded From Group
 */

public class MembersOnTransferExcludedFromGroupImpl extends BehaviourOnTransferImpl
{

  //*************************************************************************************//
  //**************************** CONSTRUCTOR ********************************************//
  //*************************************************************************************//

  public MembersOnTransferExcludedFromGroupImpl(GroupHandler groupHandler,
                                                DomainGroupHandler domainGroupHandler,
                                                DomainsUser user,
                                                SubgroupsManager subgroupsManager,
                                                PropertyManager propertyManager,
                                                DomainMessagesHandler messageHandler,
                                                int memberIdentity,
                                                View firstView)
  {
    super(groupHandler, domainGroupHandler, user, subgroupsManager, propertyManager, messageHandler, memberIdentity,
      firstView, BehaviourOnViewChanges.MembersOnTransferExcludedFromGroup);

    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> MembersOnTransferExcludedFromGroupImpl ( GroupHandler groupHandler , DomainGroupHandler domainGroupHandler , DomainsUser user , SubgroupsManager subgroupsManager , PropertyManager propertyManager , DomainMessagesHandler messageHandler , int memberIdentity , View firstView )");

    transactionsManager = new TransactionsManager(memberIdentity, groupHandler, messageHandler,
      messenger, eventsQueue, this);
  }

  //*************************************************************************************
  //**************************** FREE RESOURCES *****************************************
  //*************************************************************************************
  /**
   * Frees any resources, the object can not be used anymore
   */
  public void freeResources()
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> void freeResources ( )");
    super.freeResources();
    if (transactionsManager!=null)
    {
      transactionsManager.freeResources();
      transactionsManager=null;
    }
  }

  //*************************************************************************************
  //*************************** PROCESS ALARM EVENT *************************************
  //*************************************************************************************

  /**
   * The user of the EventsQueue can set the alarm in different ways, mainly throught the value returned
   * on the operation newViewEvent. If the alarm is activated and there are no events to send, this method
   * is called.
   */
  public void processAlarmEvent()
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> void processAlarmEvent ( )");
    Trace.processAlarmEvent();

    if (hasState())
    {
      //it must have no coordination, or, if it has coordination, there is as well a stateTransferTask working
      if (getCoordinatingMember()!=null)
        changeViewOnSenderThread(); //it there is some transfer running, the view has changed, inform of it
      else
        checkTransferToDo(); //check new coordinations
    }
    else //no state
    {
      //if it has a coordinator, there is a transfer going on, otherwise it is just checked for state assumption
      if (getCoordinator()==null)
        checkStateAssumption();
    }
  }

  //*************************************************************************************
  //*************************** CHANGING VIEW EVENT *************************************
  //*************************************************************************************

  /**
   * Event thrown after a changing view event.
   * @return CONTINUE_BLOCKING_PREVIOUS to keep blocked the previous messages
   * @return DISCARD_PREVIOUS_MESSAGES to discard any previous messages. Alarm is activated
   * @return UNBLOCK_PREVIOUS_VIEW to process messages from the previous view. Alarm is activated
   * @return UNBLOCK_ALL_PREVIOUS_VIEWS to process messages from the previous views. Alarm is activated
   */
  public int changingViewEvent()
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> int changingViewEvent ( )");
    setViewAsBlocked();

    try
    {
      subgroupsManager.changingView();
    }
    catch(SubgroupException sex)
    {
      subgroupError(sex);
    }
    //do not do anything. If they were unblocked, this call does nothing.
    //if the member has no state and there is no transfer, it should discard the messages, but that can be done
    //the same on processNewViewEvent
    return CONTINUE_BLOCKING_PREVIOUS;
  }

  //*************************************************************************************
  //*************************** PROCESS NEW VIEW EVENT **********************************
  //*************************************************************************************

  /**
   * Event thrown after a view event. The view doesn't yet have to be installed, althought the state transfer
   * system can already see the changes and return therefore the right action to do with any blocked message.
   * @return CONTINUE_BLOCKING_PREVIOUS to keep blocked the previous messages
   * @return DISCARD_PREVIOUS_MESSAGES to discard any previous messages. Alarm is activated
   * @return UNBLOCK_PREVIOUS_VIEW to process messages from the previous view. Alarm is activated
   * @return UNBLOCK_ALL_PREVIOUS_VIEWS to process messages from the previous views. Alarm is activated
   */
  public int newViewEvent(View view)
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> int newViewEvent ( View view )");
    int ret;

    //being coordinator or being coordinated, view changes are first processed
    processViewChanges(view);

    if (hasState())
      ret=newViewEventOnStatefullMembers();
    else
      ret=newViewEventOnStatelessMembers();

    Trace.newViewEvent(ret);
    return ret;
  }

  /**
   * Performs the new view event on statefull members
   */
  public int newViewEventOnStatefullMembers()
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> int newViewEventOnStatefullMembers ( )");

    int ret;
    //if there were no coordinating member, previous messages can be processed, is not blocked
    Integer coordinating = getCoordinatingMember();
    if (coordinating==null)
      ret=EventsQueueUser.UNBLOCK_PREVIOUS_VIEW;
    else
    {
      //if the old member is still up, transfer just continues
      Integer newCoordinating = getMemberToCoordinate();
      if (coordinating.equals(newCoordinating))
        ret=EventsQueueUser.CONTINUE_BLOCKING_PREVIOUS;
      else
      {
        if (isTransferStarted())
        {
          //the stateSenderThread must be completely stopped. It is obtained before the associated coordinations
          //if they are null, it means that the transfer never really started (message C not received)
          CoordinationsMap coordinationsFromStateTransferThread = stopStateSenderThread();

          //The application receives the message that the transfer has stopped, and it is included the
          //identity of the member down
          if (coordinationsFromStateTransferThread!=null)
            try
            {
              subgroupsManager.stopTransfer(coordinating.intValue(), coordinationsFromStateTransferThread, false);
              coordinationsFromStateTransferThread.freeResources();
            }
            catch(SubgroupException sex)
            {
              subgroupError(sex);
            }
        }
        setCoordinatingMember(null); //remove any reference to the coordinated member
        messageC=null;
        ret=EventsQueueUser.UNBLOCK_ALL_PREVIOUS_VIEWS;
      }
    }
    return ret;
  }

  /**
   * Performs the new view event on stateless members
   */
  public int newViewEventOnStatelessMembers()
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> int newViewEventOnStatelessMembers ( )");
    int ret;

    //if there were no coordinator, previous messages are discarded
    Integer coordinator = getCoordinator();
    if (coordinator==null)
      ret=EventsQueueUser.DISCARD_PREVIOUS_MESSAGES;
    else
    {
      //if the coordinator is still up, transfer just continues
      if (membersStateHandler.hasStateTheMember(coordinator))
        ret=EventsQueueUser.CONTINUE_BLOCKING_PREVIOUS;
      else
      {
        if (propertyManager!=null)
          propertyManager.resetGroupProperties();
        try
        {
          //transfer stops (unsuccesfully)
          subgroupsManager.stopTransfer(coordinator.intValue(), getCoordinations(false, false), false);
          removeCoordinator(true); //remove any reference to the coordinator
        }
        catch(SubgroupException sex)
        {
          subgroupError(sex);
        }
        ret=EventsQueueUser.DISCARD_PREVIOUS_MESSAGES;
      }
    }

    return ret;
  }

  //*************************************************************************************
  //**************************** PROCESS MESSAGE C **************************************
  //*************************************************************************************

  void processMessageC(int sender, MessageC message)
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> void processMessageC ( int sender , MessageC message )");
    //receiving the coordination message, it must come from the coordinated member, and it can not had been already
    //received. Additionally, this member must have already state
    if (!hasState())
      Trace.errorReceivingMessageC("member has no state");
    else
    {
      Integer coordinating=getCoordinatingMember();
      if (coordinating==null)
        Trace.errorReceivingMessageC("member has no member being coordinated");
      else if (coordinating.intValue()!=sender)
        Trace.errorReceivingMessageC("member sending the message is not the member being coordinated");
      else if (isTransferStarted())
        Trace.errorReceivingMessageC("the coordinated member has already sent the message");
      else
      {
        Trace.receivedMessageC(sender, message);

        //the transfer starts now, but only if there is no transfer going on
        synchronized(this)
        {
          if (transactionsManager.isOnTransaction())
            messageC = message;
          else
            startStateSenderThread(message, false);
        }
      }
    }
  }

  //*************************************************************************************
  //**************************** PROCESS MESSAGE E **************************************
  //*************************************************************************************

  void processMessageE(int sender, MessageE message)
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> void processMessageE ( int sender , MessageE message )");
    int newStateMember = message.member;
    Integer theSender = new Integer(sender);
    Integer theMember = new Integer(newStateMember);
    Trace.receivedMessageE(sender, newStateMember);

    if (membersStateHandler.memberGotStateFromCoordinator(theSender, theMember))
    {
      try
      {
        if (propertyManager!=null)
          propertyManager.setProperties(newStateMember, message.props, true);

        if (theSender.equals(memberId))
        {
          Debug.assert(theMember.equals(getCoordinatingMember()), Consts.AREA, "MembersOnTransferExcludedFromGroupImpl::processMessageE::1");
          CoordinationsMap coordinationsFromStateTransferThread = completeStateSenderThread();
          subgroupsManager.stopTransfer(newStateMember, coordinationsFromStateTransferThread,true);
          coordinationsFromStateTransferThread.freeResources();
          setCoordinatingMember(null); //remove the coordinator
          reportPropertiesUpdate(newStateMember);
          //unblock messages from previous views(not from this one)
          eventsQueue.unblock(true, false);
        }
        else if (theMember.equals(memberId))
        {
          CoordinationsMap coordinations=getCoordinations(false, false); //cannot be called after becameStatefull
          becameStatefull(false);
          subgroupsManager.stopTransfer(sender, coordinations,true);
          coordinations.freeResources();
          coordinations=null;
          //unblock messages from previous views, and also from this view (it will not start any transfer)
          eventsQueue.unblock(true, true);
        }
        else
          reportPropertiesUpdate(newStateMember);
      }
      catch(SubgroupException sex)
      {
        subgroupError(sex);
      }
    }
    else
      Trace.errorReceivingMessageE("member= " + sender);
  }

  //*************************************************************************************
  //**************************** PROCESS MESSAGE P **************************************
  //*************************************************************************************

  void processMessageP(int sender, MessageP message)
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> void processMessageP ( int sender , MessageP message )");
    //receiving state. Must have no state, no coordinator. If is received on a different view, is discarded
    if (message.viewId!=viewId)
      Trace.errorReceivingMessageP("view Id doesn not match the current one: "+ message.viewId + " - " + viewId);
    else if (hasState())
      Trace.errorReceivingMessageP("member has state already");
    else if (getCoordinator()!=null)
      Trace.errorReceivingMessageP("member has already a coordinator");
    else
    {
      Trace.receivedMessageP(sender, message);

      if (checkGroupFeatures(message))
      {
        //coordinator is now known, information about the group is rebuilt and the transfer starts
        Integer coordinator = new Integer(sender);
        setCoordinator(coordinator, message.replicasInfo); //information on the members is rebuilt

        Debug.assert(membersStateHandler.hasStateTheMember(coordinator), Consts.AREA, "BehaviourOnTransferImpl::processMessageP::1");

        if (propertyManager!=null)
          propertyManager.setGroupProperties(message.props);

        //inform to the subgrups that the transfer start, and send the message C. Properties must be sent, if in use
        syncTransfer(true);
      }
    }
 }

  //*************************************************************************************
  //**************************** PROCESS MESSAGE S **************************************
  //*************************************************************************************

  void processMessageS(int sender, MessageS message)
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> void processMessageS ( int sender , MessageS message )");
    //receiving the state message, it must come from the coordinating member.
    if (hasState())
      Trace.errorReceivingMessageS("member has state");
    else
    {
      Integer coordinator=getCoordinator();
      if (coordinator==null)
        Trace.errorReceivingMessageS("member has no coordinator");
      else if (coordinator.intValue()!=sender)
        Trace.errorReceivingMessageS("member sending the message is not the coordinator");
      else
      {
        Trace.receivedMessageS(sender, message);
        //information is traslated into coordinations and states
        Map subgroupStates = new HashMap();
        CoordinationsMap coordinations = setCoordinations(message.states, subgroupStates);

        if (coordinations==null)
          Trace.errorReceivingMessageS("States do not correspond with current knowledge");
        else
        {
          //states are sent to the subgroups
          try
          {
            subgroupsManager.sendState(subgroupStates, coordinations);
          }
          catch(SubgroupException ex)
          {
            subgroupError(ex);
          }
        }
      }
    }
  }

  public boolean hasStrongSTMessageConsistency(){return false;}

  //*************************************************************************************
  //**************************** TRANSACTION METHODS ************************************
  //*************************************************************************************

  public void startTransaction(Monitor monitor) throws MonitorException, MemberStateException
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> void startTransaction ( Monitor monitor )");
    transactionsManager.startTransaction(monitor);
  }

  public synchronized void endTransaction() throws MonitorException, TransactionException, MemberStateException
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> synchronized void endTransaction ( )");
    transactionsManager.endTransaction();
    if (!transactionsManager.isOnTransaction() && (messageC!=null))
    {
      startStateSenderThread(messageC, false);
      messageC=null;
    }
  }

  public void processCastMessageEvent(int sender, Message message)
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> void processCastMessageEvent ( int sender , Message message )");
    if (message instanceof EndOfTransactionMessage)
      transactionsManager.processMessageEndOfTransaction(sender, (EndOfTransactionMessage) message);
    else
      super.processCastMessageEvent(sender, message);
  }

  /**
   * Set of operartions to be performed when the member becomes statefull
   */
  void becameStatefull(boolean stateAssumed)
  {
    Trace.code("MembersOnTransferExcludedFromGroupImpl.java -> void becameStatefull ( boolean stateAssumed )");
    super.becameStatefull(stateAssumed);
    transactionsManager.becameStatefull();
  }
  //*************************************************************************************
  //**************************** DATA MEMBERS *******************************************
  //*************************************************************************************

  TransactionsManager transactionsManager;
  MessageC messageC;
}
