/**
  * File: StateMembersNotExcludedFromGroupImpl.java
  * Content: Implementation middleware-independent of the DomainGroupHandler interface with behaviour
  *          of State Members Not Excluded From Group (in fact, this includes two cases:
  *          stateless members belonging to group and stateless members excluded from group
  * Author: LuisM Pena
  * Date: 03rd October 2000
  * Version: 0.51.00
  * Last change:
  *
  *
  **/

package sensei.domains;

import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.View;
import sensei.middleware.domains.BehaviourOnViewChanges;
import sensei.middleware.domains.DomainGroupHandler;
import sensei.middleware.domains.Monitor;
import sensei.middleware.domains.MessageC;
import sensei.middleware.domains.MessageE;
import sensei.middleware.domains.MessageP;
import sensei.middleware.domains.MessageS;
import sensei.middleware.domains.TransactionException;

import sensei.util.Debug;

import java.util.HashMap;
import java.util.Map;

/**
 * Implementation middleware-independent of the DomainGroupHandler interface with behaviour
 * of State Members Not Excluded From Group (in fact, this includes two cases:
 * stateless members belonging to group and stateless members excluded from group
 */

public class StateMembersNotExcludedFromGroupImpl extends BehaviourOnTransferImpl
{

  //*************************************************************************************//
  //**************************** CONSTRUCTOR ********************************************//
  //*************************************************************************************//

  public StateMembersNotExcludedFromGroupImpl(  GroupHandler groupHandler,
                                                DomainGroupHandler domainGroupHandler,
                                                DomainsUser user,
                                                SubgroupsManager subgroupsManager,
                                                PropertyManager propertyManager,
                                                DomainMessagesHandler messageHandler,
                                                int memberIdentity,
                                                View firstView,
                                                boolean statelessMembersExcludedFromGroup)
  {
    super(groupHandler, domainGroupHandler, user, subgroupsManager, propertyManager, messageHandler, memberIdentity,
      firstView, statelessMembersExcludedFromGroup? BehaviourOnViewChanges.StatelessMembersDoNotBelongToGroup :
      BehaviourOnViewChanges.StatelessMembersBelongToGroup);

    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> StateMembersNotExcludedFromGroupImpl ( GroupHandler groupHandler , DomainGroupHandler domainGroupHandler , DomainsUser user , SubgroupsManager subgroupsManager , PropertyManager propertyManager , DomainMessagesHandler messageHandler , int memberIdentity , View firstView , boolean statelessMembersExcludedFromGroup )");

    this.statelessMembersExcludedFromGroup = statelessMembersExcludedFromGroup;
    syncDone=false;
    propertiesMustBeSent=true;
  }

  //*************************************************************************************
  //*************************** PROCESS ALARM EVENT *************************************
  //*************************************************************************************

  /**
   * The user of the EventsQueue can set the alarm in different ways, mainly throught the value returned
   * on the operation newViewEvent. If the alarm is activated and there are no events to send, this method
   * is called.
   */
  public void processAlarmEvent()
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> void processAlarmEvent ( )");
    Trace.processAlarmEvent();

    if (hasState())
    {
      //if there is a transfer going on, it must wait for the message C, otherwise it is checked for new transfers.
      if (getCoordinatingMember()==null)
        checkTransferToDo(); //check new coordinations, unblocking messages if no members to coordinate
    }
    else //no state
    {
      //if it has a coordinator, there is a transfer going on, and it must synchronize the transfer.
      //Otherwise, it is needed to check if the member must assume state
      Integer coordinator=getCoordinator();
      if (coordinator==null)
        checkStateAssumption();
      else
        syncTransfer(propertiesMustBeSent); //properties are not sent, except if they have changed
    }
  }

  /**
   * Override from the super class, to verify on the client side when the transfer is going on
   */
  void syncTransfer(boolean sendProperties)
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> void syncTransfer ( boolean sendProperties )");
    syncDone=true;
    propertiesMustBeSent=false; //already sent!
    super.syncTransfer(sendProperties);
  }

  //*************************************************************************************
  //*************************** CHANGING VIEW EVENT *************************************
  //*************************************************************************************

  /**
   * Event thrown after a changing view event.
   * @return CONTINUE_BLOCKING_PREVIOUS to keep blocked the previous messages
   * @return DISCARD_PREVIOUS_MESSAGES to discard any previous messages. Alarm is activated
   * @return UNBLOCK_PREVIOUS_VIEW to process messages from the previous view. Alarm is activated
   * @return UNBLOCK_ALL_PREVIOUS_VIEWS to process messages from the previous views. Alarm is activated
   */
  public int changingViewEvent()
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> int changingViewEvent ( )");
    setViewAsBlocked();
    //interrupt first any transfer, if there is some transfer on execution
    pauseStateSenderThread();

    try
    {
      subgroupsManager.changingView();
      if (hasState())
        interruptStateSenderThread();
      else if (syncDone)
      {
        syncDone=false;
        //if some synchronization was done, interrupt it
        subgroupsManager.interruptTransfer(getCoordinations(false, false));
      }
    }
    catch(SubgroupException sex)
    {
      subgroupError(sex);
    }
    //a member with state must return unblockView, a member withou state will return the same
    //when the stateless members belong to the group, or DISCARD if they do not belong
    return !statelessMembersExcludedFromGroup || hasState()? UNBLOCK_PREVIOUS_VIEW : DISCARD_PREVIOUS_MESSAGES;
  }

  //*************************************************************************************
  //*************************** PROCESS NEW VIEW EVENT **********************************
  //*************************************************************************************

  /**
   * Event thrown after a view event. The view doesn't yet have to be installed, althought the state transfer
   * system can already see the changes and return therefore the right action to do with any blocked message.
   * @return CONTINUE_BLOCKING_PREVIOUS to keep blocked the previous messages
   * @return DISCARD_PREVIOUS_MESSAGES to discard any previous messages. Alarm is activated
   * @return UNBLOCK_PREVIOUS_VIEW to process messages from the previous view. Alarm is activated
   * @return UNBLOCK_ALL_PREVIOUS_VIEWS to process messages from the previous views. Alarm is activated
   */
  public int newViewEvent(View view)
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> int newViewEvent ( View view )");
    int ret;

    //being coordinator or being coordinated, view changes are first processed
    processViewChanges(view);

    if (hasState())
      ret=newViewEventOnStatefullMembers();
    else
      ret=newViewEventOnStatelessMembers();

    Trace.newViewEvent(ret);

    return ret;
  }

  /**
   * Performs the new view event on statefull members
   */
  public int newViewEventOnStatefullMembers()
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> int newViewEventOnStatefullMembers ( )");
    //if there were any coordination, it must be stopped
    Integer coordinating = getCoordinatingMember();
    if (coordinating!=null)
    {
      //the message to stop the transfer must include the identity of the member down. It is checked first
      //if the coordinated member is down
      Integer newCoordinating = getMemberToCoordinate();
      boolean coordinatedDown = !coordinating.equals(newCoordinating);

      if (coordinatedDown)
      {
        //stop the sender thread and communicate the stop to the application. This is only needed if the
        //transfer had really started
        if (isTransferStarted())
        {
          CoordinationsMap coordinationsFromStateTransferThread = stopStateSenderThread();
          try
          {
            subgroupsManager.stopTransfer(coordinating.intValue(), coordinationsFromStateTransferThread, false);
          }
          catch(SubgroupException sex)
          {
            subgroupError(sex);
          }
          coordinationsFromStateTransferThread.freeResources();
        }
        //Remove any reference to the coordinator
        setCoordinatingMember(null);
        propertiesMustBeSent=true;
        syncDone=false;
      }
    }
    //having state, messages from the previous view must be always processed
    return EventsQueueUser.UNBLOCK_PREVIOUS_VIEW;
  }

  /**
   * Performs the new view event on stateless members
   */
  public int newViewEventOnStatelessMembers()
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> int newViewEventOnStatelessMembers ( )");
    //if there were a coordinator, the transfer is stopped
    Integer coordinator = getCoordinator();
    if (coordinator!=null)
    {
      //if the coordinator is down, stop must be communicated to the application, and any group properties reset
      if (!membersStateHandler.hasStateTheMember(coordinator))
      {
        if (propertyManager!=null)
          propertyManager.resetGroupProperties();
        try
        {
          //transfer stops (unsuccesfully)
          subgroupsManager.stopTransfer(coordinator.intValue(), getCoordinations(false, false), false);
          removeCoordinator(true); //remove any reference to the coordinator
        }
        catch(SubgroupException sex)
        {
          subgroupError(sex);
        }
      }
    }
    //if new members belong to the group, previous messages must be processed, otherwise discarded
    return statelessMembersExcludedFromGroup?
        EventsQueueUser.DISCARD_PREVIOUS_MESSAGES :
        EventsQueueUser.UNBLOCK_PREVIOUS_VIEW;
  }

  //*************************************************************************************
  //**************************** PROCESS MESSAGE C **************************************
  //*************************************************************************************
  void processMessageC(int sender, MessageC message)
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> void processMessageC ( int sender , MessageC message )");
    //receiving the coordination message, it must come from the coordinated member
    //if it's the first time that the coordinated sends a message C, it must contain its properties, next messages
    //do not include properties. A member can send several messages C, but only once per view. If it does not come
    //on the same view is directly not used. Additionally, this member must have already state
    if (!hasState())
      Trace.errorReceivingMessageC("member has no state");
    else
    {
      Integer coordinating=getCoordinatingMember();
      if (coordinating==null)
        Trace.errorReceivingMessageC("member has no member being coordinated");
      else if (coordinating.intValue()!=sender)
        Trace.errorReceivingMessageC("member sending the message is not the member being coordinated");
      else if (viewId != message.viewId)
        Trace.errorReceivingMessageC("message received on a different view");
      else if (message.containOwnProperties && (propertyManager==null))
        Trace.errorReceivingMessageC("message contains properties and the group does not use them");
      else if (isViewAlreadyBlocked())
        Trace.errorReceivingMessageC("message received on the blocked view, discarded");
      else
      {
        Trace.receivedMessageC(sender, message);
        if (isTransferStarted())
          continueStateSenderThread(message);
        else
          startStateSenderThread(message, true);
      }
    }
  }

  //*************************************************************************************
  //**************************** PROCESS MESSAGE E **************************************
  //*************************************************************************************
  void processMessageE(int sender, MessageE message)
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> void processMessageE ( int sender , MessageE message )");
    int newStateMember = message.member;
    Integer theSender = new Integer(sender);
    Integer theMember = new Integer(newStateMember);
    Trace.receivedMessageE(sender, newStateMember);

    //message must be received on the same view
    if (viewId!=message.viewId)
      Trace.errorReceivingMessageE("coming on a different view");
    else if (isViewAlreadyBlocked())
      Trace.errorReceivingMessageE("coming on a blocked view");
    else if (membersStateHandler.memberGotStateFromCoordinator(theSender, theMember))
    {
      try
      {
        if (propertyManager!=null)
          propertyManager.setProperties(newStateMember, message.props, true);

        if (theSender.equals(memberId))
        {
          Debug.assert(theMember.equals(getCoordinatingMember()), Consts.AREA, "StateMembersNotExcludedFromGroupImpl::processMessageE::1");
          CoordinationsMap coordinationsFromStateTransferThread = completeStateSenderThread();
          subgroupsManager.stopTransfer(newStateMember, coordinationsFromStateTransferThread,true);
          coordinationsFromStateTransferThread.freeResources();
          setCoordinatingMember(null); //remove the coordinator
          reportPropertiesUpdate(newStateMember);
          //start probably new transfers
          checkTransferToDo();
        }
        else if (theMember.equals(memberId))
        {
          CoordinationsMap coordinations=getCoordinations(false, false); //cannot be called after becameStatefull
          becameStatefull(false);
          subgroupsManager.stopTransfer(sender, coordinations,true);
          coordinations.freeResources();
          coordinations=null;
          //unblock messages from this view and previous views
          eventsQueue.unblock(true, true);
        }
        else
          reportPropertiesUpdate(newStateMember);
      }
      catch(SubgroupException sex)
      {
        subgroupError(sex);
      }
    }
    else
    {
      Trace.errorReceivingMessageE("member= " + sender);
      Debug.assert(!theSender.equals(memberId), Consts.AREA,"StateMembersNotExcludedFromGroupImpl::processMessageE");
    }
  }

  //*************************************************************************************
  //**************************** PROCESS MESSAGE P **************************************
  //*************************************************************************************

  void processMessageP(int sender, MessageP message)
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> void processMessageP ( int sender , MessageP message )");
    //receiving state. Must have no state, no coordinator. If is received on a different view, is discarded
    if (message.viewId!=viewId)
      Trace.errorReceivingMessageP("view Id doesn not match the current one: "+ message.viewId + " - " + viewId);
    else if (hasState())
      Trace.errorReceivingMessageP("member has state already");
    else if (getCoordinator()!=null)
      Trace.errorReceivingMessageP("member has already a coordinator");
    else
    {
      Trace.receivedMessageP(sender, message);

      if (checkGroupFeatures(message))
      {
        //coordinator is now known, information about the group is rebuilt and the transfer starts
        Integer coordinator = new Integer(sender);
        setCoordinator(coordinator, message.replicasInfo); //information on the members is rebuilt

        Debug.assert(membersStateHandler.hasStateTheMember(coordinator), Consts.AREA, "BehaviourOnTransferImpl::processMessageP::1");

        if (propertyManager!=null)
          propertyManager.setGroupProperties(message.props);

        //inform to the subgrups that the transfer start, and send the message C. Properties must be sent, if in use
        //if the message P has been received after the changeViewEvent, it is not needed to send the message
        if (!isViewAlreadyBlocked())
          syncTransfer(true);
      }
    }
 }

  //*************************************************************************************
  //**************************** PROCESS MESSAGE S **************************************
  //*************************************************************************************

  void processMessageS(int sender, MessageS message)
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> void processMessageS ( int sender , MessageS message )");
    //receiving the state message, it must come from the coordinating member.
    if (hasState())
      Trace.errorReceivingMessageS("member has state");
    else
    {
      Integer coordinator=getCoordinator();
      if (coordinator==null)
        Trace.errorReceivingMessageS("member has no coordinator");
      else if (coordinator.intValue()!=sender)
        Trace.errorReceivingMessageS("member sending the message is not the coordinator");
      else if (viewId!=message.viewId)
        Trace.errorReceivingMessageS("it has arrived on a different view");
      else if (isViewAlreadyBlocked())
        Trace.errorReceivingMessageS("it has arrived on a blocked view");
      else
      {
        Trace.receivedMessageS(sender, message);
        //information is traslated into coordinations and states
        Map subgroupStates = new HashMap();
        CoordinationsMap coordinations = setCoordinations(message.states, subgroupStates);

        if (coordinations==null)
          Trace.errorReceivingMessageS("States do not correspond with current knowledge");
        else
        {
          //states are sent to the subgroups
          try
          {
            subgroupsManager.sendState(subgroupStates, coordinations);
          }
          catch(SubgroupException ex)
          {
            subgroupError(ex);
          }
        }
      }
    }
  }

  public boolean hasStrongSTMessageConsistency(){return true;}

  //*************************************************************************************
  //**************************** TRANSACTION METHODS ************************************
  //*************************************************************************************

  public void startTransaction(Monitor monitor) throws TransactionException
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> void startTransaction ( Monitor monitor )");
    //not allowed
    throw new TransactionException();
  }

  public void endTransaction() throws TransactionException
  {
    Trace.code("StateMembersNotExcludedFromGroupImpl.java -> void endTransaction ( )");
    //not allowed
    throw new TransactionException();
  }

  //*************************************************************************************
  //**************************** DATA MEMBERS *******************************************
  //*************************************************************************************

  boolean statelessMembersExcludedFromGroup;
  boolean propertiesMustBeSent;
  //set to true if the member has done some synhronization this view and it is suitable to be interrupted
  boolean syncDone;
}
