/**
  * File: StateSenderThread.java
  * Content: Separate thread class to do the state transfer
  * Author: LuisM Pena
  * Date: 2nd August 2001
  * Version: 0.51.00
  * Last change:
  *
  *
  **/

package sensei.domains;

import sensei.middleware.domains.PhaseCoordination;
import sensei.middleware.domains.MemberProperties;
import sensei.middleware.domains.Property;
import sensei.middleware.domains.State;
import sensei.middleware.domains.SubgroupCoordination;
import sensei.middleware.domains.SubgroupState;

import sensei.util.Debug;
import sensei.util.Error;
import sensei.util.IntWrapper;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Separate thread class to do the state transfer
 */
class StateSenderThread extends Thread
{
  /**
   * The constructor requires several parameters to proceed with the transfer:
   * @param target: the member receiving the transfer
   * @param coordinations: the coordinations sent by the client on a syncTransfer method
   * @param targetProperties: properties associated to the target. Cannot be null
   * @param manager: the subgroupsManager required to interact with the subgroups
   * @param messenger: the object to send the messages
   * @param user: receives the results of the state transfer
   * @param initialViewId
   * @param startPaused: to true if the transfer does not start immediately
   */
  public StateSenderThread(int target, SubgroupCoordination[] clientCoordinations, SubgroupsManager subgroupsManager,
          Property[] targetProperties, DomainsFactory messenger, StateSenderThreadUser user,
          int initialViewId, boolean startPaused)
  {
    Trace.code("StateSenderThread.java -> StateSenderThread ( int target , SubgroupCoordination [ ] clientCoordinations , SubgroupsManager subgroupsManager , Property [ ] targetProperties , DomainsFactory messenger , StateSenderThreadUser user , int initialViewId , boolean startPaused )");
    this.target=target;
    this.subgroupsManager=subgroupsManager;
    this.messenger=messenger;
    this.user=user;
    this.coordinations=convertToCoordinationsMap(clientCoordinations);
    this.targetProperties = targetProperties;
    viewId=initialViewId;
    stoppedReceived=true;
    pausedReceived=true;
    callInterrupt=false;
    discardMessagesSReady=false;
    status=startPaused? PAUSED : TRANSFERING;
    start();
  }

  //*************************************************************************************
  //**************************** FREE RESOURCES *****************************************
  //*************************************************************************************
  /**
   * Frees any resources, the object cannot be used anymore
   */
  public synchronized void freeResources()
  {
    Trace.code("StateSenderThread.java -> synchronized void freeResources ( )");
    if ((status!=ENDED)||(status!=STOPPED))
      stopTransfer();
  }

  //*************************************************************************************
  //**************************** GET COORDINATIONS **************************************
  //*************************************************************************************

  public CoordinationsMap getCoordinations()
  {
    Trace.code("StateSenderThread.java -> CoordinationsMap getCoordinations ( )");
    return coordinations;
  }

  //*************************************************************************************
  //**************************** THREAD METHOD ******************************************
  //*************************************************************************************

  public void run()
  {
    Trace.code("StateSenderThread.java -> void run ( )");
    if (checkStateTransfer())
    {
      try
      {
        callInterrupt = true;
        subgroupsManager.startTransfer(target, coordinations);

        //send sequentially the state messages
        while(checkStateTransfer())
        {
          if (coordinations.anyValidCoordination) //send still the state
          {
            discardMessagesSReady=false;
            Map states=subgroupsManager.getState(coordinations);
            if (checkStateTransfer() && !discardMessagesSReady)
            {
              Trace.sendMessageS(target);
              if (!messenger.sendMessageS(target, viewId, getSubgroupStates(coordinations, states))
                  && user.hasStrongSTMessageConsistency())
                status=PAUSED; //it is sent on a different view, or directly lost
            }
          }
          else
          {
            //state sent: send the message E: it must be received on the same view, before the blocking in some
            //cases (in MembersExcludedFromGroup, this is not needed!)
            Trace.castMessageE(target);
            messageEisReceived=false;
            boolean sentOnSameView = messenger.castMessageE(viewId, target, targetProperties);
            if (user.hasStrongSTMessageConsistency())
              if (sentOnSameView)
                checkEndTransfer();
              else
                status=PAUSED;
            else
              synchronized(this)
              {
                status = ENDED;
                notifyAll();
              }
          }
        }
      }
      catch(SubgroupException ex)
      {
        user.stateTransferThreadSubgroupError(ex);
      }
      catch(Exception ex)
      {
        Error.unhandledException(Consts.AREA, ex); //Remote/CORBA exception, menaingless as the access is local!!!
      }
    }
    subgroupsManager=null;
    messenger=null;
    targetProperties=null;
    user=null;
//    It is not possible here
//    coordinations.freeResources();
//    coordinations=null;
    Trace.code("StateSenderThread.java -> void run ( ) ended");
  }

  //*************************************************************************************
  //**************************** MESSAGE E RECEIVED *************************************
  //*************************************************************************************

  /**
   * Confirms that the sent E message has been received, and successfully
   */
  public synchronized void completedTransfer()
  {
    Trace.code("StateSenderThread.java -> synchronized void completedTransfer ( )");
    messageEisReceived = true;
    notifyAll();
  }

  //*************************************************************************************
  //**************************** PAUSE TRANSFER *****************************************
  //*************************************************************************************

  /**
   * Stops temporarily the transfer; do not come back until the transfer stops. It can be
   * continued with resumeTransfer or stopped definitely with stopTransfer
   */
  public synchronized void pauseTransfer()
  {
    Trace.code("StateSenderThread.java -> synchronized void pauseTransfer ( )");
    if ((status!=ENDED) && (status!=PAUSED))
    {
      Debug.assert(pausedReceived,Consts.AREA, "StateSenderThread::pauseTransfer::2");
      Debug.assert(status!=STOPPED,Consts.AREA, "StateSenderThread::pauseTransfer::3");
      status=PAUSED;
      pausedReceived=false;
      notifyAll();
      while(status!=ENDED && !pausedReceived)
        try{wait();}catch(InterruptedException ex){}
    }
    Trace.code("StateSenderThread.java -> -- synchronized void pauseTransfer ( )");
  }

  //*************************************************************************************
  //**************************** STOP TRANSFER ******************************************
  //*************************************************************************************

  /**
   * Stops completely the transfer; do not come back until the transfer ends.
   */
  public synchronized void stopTransfer()
  {
    Trace.code("StateSenderThread.java -> synchronized void stopTransfer ( )");
    if (status!=ENDED)
    {
      Debug.assert(stoppedReceived,Consts.AREA, "StateSenderThread::stopTransfer::2");
      status=STOPPED;
      stoppedReceived=false;
      notifyAll();
      while(status!=ENDED && !stoppedReceived)
        try{wait();}catch(InterruptedException ex){}
    }
    Trace.code("StateSenderThread.java -> -- synchronized void stopTransfer ( )");
  }

  //*************************************************************************************
  //**************************** INTERRUPT TRANSFER *************************************
  //*************************************************************************************

  /**
   * Interrupts the transfer, which must be paused. After a call to interrupt, only a
   * call to continue or another call to pause are allowed.
   */
  public synchronized void interruptTransfer() throws SubgroupException
  {
    Trace.code("StateSenderThread.java -> synchronized void interruptTransfer ( )");
    Debug.assert(status==PAUSED || status==ENDED,Consts.AREA, "StateSenderThread::interruptTransfer::1");
    if (callInterrupt)
    {
      subgroupsManager.interruptTransfer(coordinations);
      callInterrupt=false;
    }
    if (status!=ENDED)
      status=INTERRUPTED;
  }

  //*************************************************************************************
  //**************************** CONTINUE TRANSFER **************************************
  //*************************************************************************************

  /**
   * Continues an interrupted transfer. It needs the coordinations, given by the joining member in a message C
   * @param newViewId The transfer is paused between views. When it is resumed, the new view Id must be given
   * @param clientCoordinations The coordinations given by the client
   * @param clientProperties The client properties, if have changed. If not, set it to null
   */
  public synchronized void continueTransfer(SubgroupCoordination[] clientCoordinations,
    Property[] newClientProperties)
      throws SubgroupException
  {
    Trace.code("StateSenderThread.java -> synchronized void continueTransfer ( SubgroupCoordination [ ] clientCoordinations , Property [ ] newClientProperties )");
    Debug.assert(status==INTERRUPTED,Consts.AREA, "StateSenderThread::continueTransfer::1");

    if (newClientProperties!=null)
      targetProperties = newClientProperties;

    coordinations=convertToCoordinationsMap(clientCoordinations);

    subgroupsManager.continueTransfer(target, coordinations,
                              convertToCoordinationsMap(clientCoordinations).coordinationsMap);

    callInterrupt=true;
    discardMessagesSReady=true; //any messageS that would be ready must not be sent!
    status=TRANSFERING;
    notifyAll();
  }

  //*************************************************************************************
  //**************************** CHANGE VIEW ID *****************************************
  //*************************************************************************************

  /**
   * Changes the view Id used to send the messages
   */
  public void changeViewId(int newViewId)
  {
    Trace.code("StateSenderThread.java -> void changeViewId ( int newViewId )");
    viewId=newViewId;
  }

  //*************************************************************************************
  //**************************** CHECK STATE TRANSFER ***********************************
  //*************************************************************************************

  /**
   * blocks if the transfer is paused, and returns false if the transfer is stopped
   */
  synchronized boolean checkStateTransfer()
  {
    Trace.code("StateSenderThread.java -> synchronized boolean checkStateTransfer ( )");
    if (status==PAUSED)
    {
      pausedReceived=true;
      notifyAll(); //wakes up the pause transfer method
      while((status!=TRANSFERING) && (status!=STOPPED))
        try{wait();}catch(InterruptedException ex){}
    }
    boolean ret=status!=STOPPED;
    if (!ret)
    {
      stoppedReceived=true; //wake up the stop method
      notifyAll();
    }
    Trace.code("StateSenderThread.java -> -- synchronized boolean checkStateTransfer ( )");
    return ret && (status!=ENDED);
  }

  //*************************************************************************************
  //**************************** CHECK END TRANSFER *************************************
  //*************************************************************************************

  /**
   * blocks until the user reports to have received in order the message E or until any
   * other method (PAUSE / STOP, etc) is received.
   */
  synchronized void checkEndTransfer()
  {
    Trace.code("StateSenderThread.java -> synchronized void checkEndTransfer ( )");
    while ((!messageEisReceived) && (status==TRANSFERING))
      try{wait();}catch(InterruptedException ex){}

    if (messageEisReceived)
      status=ENDED;
    Trace.code("StateSenderThread.java -> -- synchronized void checkEndTransfer ( )");
  }

  //*************************************************************************************
  //**************************** GET SUBGROUP STATES ************************************
  //*************************************************************************************

  /**
   * This method uses the coordinations and the Map with states (IntWrapper, State) to build the
   * SubgroupState[] required to be sent to the transfer member.
   */
  SubgroupState[] getSubgroupStates(CoordinationsMap coordinations, Map states)
  {
    Trace.code("StateSenderThread.java -> SubgroupState [ ] getSubgroupStates ( CoordinationsMap coordinations , Map states )");
    int size=states.size();
    SubgroupState[] ret = new SubgroupState[size];
    Iterator it = states.entrySet().iterator();
    while(it.hasNext())
    {
      Map.Entry entry = (Map.Entry) (it.next());
      IntWrapper subgroup = (IntWrapper)(entry.getKey());
      State state = (State)(entry.getValue());
      PhaseCoordination phase = ((SubgroupCoordination)(coordinations.coordinationsMap.get(subgroup))).phase;
      Debug.assert(state!=null, Consts.AREA, "StateSenderThread::getSubgroupStates::1");
      Debug.assert(phase!=null, Consts.AREA, "StateSenderThread::getSubgroupStates::2");
      ret[--size]=new SubgroupState(subgroup.i, phase, state);
    }
    Debug.assert(size==0, Consts.AREA, "StateSenderThread::getSubgroupStates::3");
    return ret;
  }

  //*************************************************************************************
  //**************************** CONVERT TO COORDINATIONSMAP ****************************
  //*************************************************************************************
  /**
   * Converts from SubgroupCoordination[] into a CoordinationsMap structure
   */
  CoordinationsMap convertToCoordinationsMap(SubgroupCoordination[] clientCoordinations)
  {
    Trace.code("StateSenderThread.java -> CoordinationsMap convertToCoordinationsMap ( SubgroupCoordination [ ] clientCoordinations )");
    int size=clientCoordinations.length;
    Map ret = new HashMap(size);
    for(int i=0;i<size;i++)
    {
      SubgroupCoordination coordination=clientCoordinations[i];
      ret.put(new IntWrapper(coordination.id), coordination);
    }
    return new CoordinationsMap(ret);
  }

  //*************************************************************************************
  //**************************** DATA MEMBERS *******************************************
  //*************************************************************************************

  final static int STOPPED=0;
  final static int PAUSED=1;
  final static int TRANSFERING=2;
  final static int INTERRUPTED=3;
  final static int ENDED=3;
  int status;
  boolean stoppedReceived, pausedReceived, messageEisReceived;
  boolean callInterrupt;

  boolean discardMessagesSReady;

  //identity of the member receiving the state
  int target;

  //identity of the current view
  int viewId;

  //subgroupsManager is used to make calls on the subgroups
  SubgroupsManager subgroupsManager;

  //The messenger is used to send DomainGroupHandler messages
  DomainsFactory messenger;

  //Coordinations for the current transfer
  CoordinationsMap coordinations;

  //target's properties
  Property[] targetProperties;

  StateSenderThreadUser user;
}
