/**
  * File: Trace.java
  * Content: Class to trace the UMA events
  * Author: LuisM Pena
  * Date: 19th October 2000
  * Last change:
  * Version: 0.51.00
  *
  * Comments: This class is not intended to be included in the release
  *						version.
  **/

package sensei.domains;

import sensei.util.Error;
import sensei.util.logging.Logger;

import sensei.middleware.domains.DomainExpulsionReason;
import sensei.middleware.domains.DomainMessage;
import sensei.middleware.domains.MessageC;
import sensei.middleware.domains.MessageP;
import sensei.middleware.domains.MessageS;
import sensei.middleware.domains.MessageE;
import sensei.middleware.domains.MessageMemberProperties;
import sensei.middleware.domains.ReplicaInformation;
import sensei.middleware.domains.StateTransferType;

/**
  * Class to trace the UMA events
  **/
public class Trace
{
  //***************************************************************************
  //*****************  EVENTS QUEUE TRACES ************************************
  //***************************************************************************

  public static void queueUnblock(boolean previousView, boolean thisView)
  {
    if (TRACE)
      print("Unblocking EventsQueue ( " +
        (previousView? "-previous view- " : "") +  (thisView? "-current view- " : "") + ")");
  }

  public static void queueUnblocked(boolean alarm, boolean currentQueued)
  {
    if (TRACE)
      print("EventsQueue unblocked, alarm is set (" + alarm + "), current messages are queued (" + currentQueued + ")");
  }

  public static void queueQ2Dumped(int events)
  {
    if (TRACE)
      print("EventsQueue: Dumping Q2 (previous views) into the events list. Q2 contains " + events + " lists");
  }

  public static void queueDumped(int events)
  {
    if (TRACE)
      print("EventsQueue: Dumping " + events + " events into the events list (unblocked)");
  }

  public static void queueExpulsionEventArrived()
  {
    if (TRACE)
      print("EventsQueue: received expulsion event, added to the queue");
  }

  public static void queueChangingViewEventArrived()
  {
    if (TRACE)
      print("EventsQueue: received new changing view event");
  }

  public static void queueNewViewEventArrived()
  {
    if (TRACE)
      print("EventsQueue: received new view event, next messages queued into new queue");
  }

  public static void queueDiscardingNewMessage()
  {
    if (TRACE)
      print("EventsQueue: discarding incoming message");
  }

  public static void queueEmptyEventList(boolean alarm)
  {
    if (TRACE)
      if (alarm)
        print("EventsQueue: event list empty. Alarm was set, the user will be called immediately");
      else
        print("EventsQueue: event list empty. Alarm was not set");
  }

  public static void queueProcessStopping()
  {
    if (TRACE)
      print("EventsQueue: no more events to process (expulsion processed). Quitting....");
  }

  public static void queueProcessEvents()
  {
    if (TRACE)
      print("EventsQueue: processing events");
  }

  public static void queueProcessEventsEnded()
  {
    if (TRACE)
      print("EventsQueue: processing events finished.");
  }

  public static void queueSendsMessage(boolean ptp)
  {
    if (TRACE)
      print("EventsQueue: sending message to the application " + (ptp? "(PTP)" : "(cast)"));
  }

  public static void queueProcessesView(boolean validView, boolean alreadyPreprocessed)
  {
    if (TRACE)
      if (!validView)
        print("EventsQueue: sending expulsion event to the application");
      else if (alreadyPreprocessed)
        print("EventsQueue: sending view to the application");
      else
        print("EventsQueue: sending view to be preprocessed to the application");
  }

  public static void queuePreprocessedView(int answer)
  {
    if (TRACE)
      print("EventsQueue: view preprocessed, answer is " +
        (answer==EventsQueueUser.CONTINUE_BLOCKING_PREVIOUS? "to continue blocking previous views" :
         answer==EventsQueueUser.DISCARD_PREVIOUS_MESSAGES?  "to discard previous messages" :
         answer==EventsQueueUser.UNBLOCK_ALL_PREVIOUS_VIEWS? "to unblock any message from previous views" :
                                                             "to unblock the messages from previous view") );
  }

  //***************************************************************************//
  //*****************  DOMAIN GROUP MEMBER IMPL TRACES ************************//
  //***************************************************************************//

  public static void memberAccepted()
  {
    if (TRACE)
      print("Member accepted");
  }

  public static void assumingState()
  {
    if (TRACE)
      print("Assuming state");
  }

  public static void assumingState(Integer member)
  {
    if (TRACE)
      print("The member " + member + " should assume state");
  }

  public static void memberToCoordinate(Integer coordinating)
  {
    if (TRACE)
      if (coordinating==null)
        print("No members to be coordinated");
      else
        print("Coordinating member " + coordinating);
  }

  public static void processAlarmEvent()
  {
    if (TRACE)
      print("Received alarm event from the EventsQueue");
  }

  public static void newViewEvent(int ret)
  {
    if (TRACE)
      switch(ret)
      {
        case EventsQueueUser.CONTINUE_BLOCKING_PREVIOUS:
          print("Received new view event. Messages kept blocked");
          break;
        case EventsQueueUser.DISCARD_PREVIOUS_MESSAGES:
          print("Received new view event. Messages are discarded");
          break;
        case EventsQueueUser.UNBLOCK_ALL_PREVIOUS_VIEWS:
          print("Received new view event. Messages are unblocked, from any view");
          break;
        case EventsQueueUser.UNBLOCK_PREVIOUS_VIEW:
          print("Received new view event. Messages from previous view are unblocked");
          break;
        default:
          print("Received new view event. WRONG ANSWER!!!!");
          break;
      }
  }

  public static void sendMessageP(int target)
  {
    if (TRACE)
      print("Sending message P to member " + target);
  }

  public static void errorReceivingMessageP(String reason)
  {
    if (TRACE_WARNINGS)
      print("Error, a message P has been received: " + reason);
  }

  static void receivedMessageP(int source, MessageP message)
  {
    if (TRACE)
      print("Received message P from " + source + ". " +  getReplicasInformation(message.replicasInfo));
  }

  public static void sendMessageC(int target)
  {
    if (TRACE)
      print("Sending message C to member " + target);
  }

  static void receivedMessageC(int source, MessageC message)
  {
    if (TRACE)
      print("Received message C from " + source);
  }

  public static void errorReceivingMessageC(String reason)
  {
    if (TRACE_WARNINGS)
      print("Error, a message C has been received: " + reason);
  }

  public static void castMessageMemberProperties()
  {
    if (TRACE)
      print("Casting message MemberProperties");
  }

  static void receivedMessageMemberProperties(int source, MessageMemberProperties message)
  {
    if (TRACE)
      print("Received message MemberProperties from " + source);
  }

  public static void errorReceivingMessageMemberProperties(String reason)
  {
    if (TRACE_WARNINGS)
      print("Error, a message MemberProperties has been received: " + reason);
  }

  public static void sendMessageS(int target)
  {
    if (TRACE)
      print("Sending message S to member " + target);
  }

  public static void errorReceivingMessageS(String reason)
  {
    if (TRACE_WARNINGS)
      print("Error, a message S has been received: " + reason);
  }

  public static void receivedMessageS(int source, MessageS message)
  {
    if (TRACE)
      print("Received message S from " + source);
  }

  public static void castMessageE(int newStateMember)
  {
    if (TRACE)
      print("Casting message E to publish new state for member " + newStateMember);
  }

  public static void receivedMessageE(int source, int member)
  {
    if (TRACE)
      print("Received message E from " + source + " to make member " + member + " as statefull");
  }

  public static void errorReceivingMessageE(String reason)
  {
    if (TRACE_WARNINGS)
      print("Error, a message E has been received and is not valid, " + reason);
  }

  public static void errorReceivingMessageDynamicBehaviour()
  {
    if (TRACE_WARNINGS)
      print("Error, dynamic subgroup's message received, but not accepted");
  }

  public static void errorCreatingDynamicSubgroup(StateTransferType expected, StateTransferType received)
  {
    if (TRACE)
      print("Error creating a dynamic subgroup. Type expected was " + toString(expected) +
          " but it was received "+toString(received));
  }

  static String toString(StateTransferType type)
  {
    switch(type.value())
    {
      case 0: return "State Handler";
      case 1: return "Basic State Handler";
      case 2: return "Checkpointable";
    }
    return "No state transfer support";
  }

  public static void checkedGroupFeatures(DomainExpulsionReason reason)
  {
    if (TRACE)
    {
      String cause=reason==DomainExpulsionReason.WrongBehaviourMode? "wrong behaviour mode" :
      reason==DomainExpulsionReason.WrongPropertyAllowance? "properties policy collision" :
        reason==DomainExpulsionReason.WrongStaticSubgroupsComposition? "static subgroups composition collision":
        reason==DomainExpulsionReason.WrongCoordinatorElectionPolicy? "coordinator election policy collision": "???";
      print("Member's features are not compliant with group features: " + cause);
    }
  }

  static String getReplicasInformation(ReplicaInformation replicas[])
  {
    int len=replicas.length;
    if (len==0)
      return("No state replicas????");
    else
    {
      StringBuffer toPrint=new StringBuffer("Replicas with state (coordinations):");
      for (int i=0;i<len;i++)
      {
        if (i>0)
          toPrint.append(", ");
        toPrint.append(replicas[i].replicaId);
        int []coordinating = replicas[i].replicasToCoordinate;
        int len2 = coordinating.length;
        if (len2>0)
        {
          toPrint.append(" (" + coordinating[0]);
          for(int j=1;j<len2;j++)
            toPrint.append(", " + coordinating[j]);
          toPrint.append(")");
        }
      }
      return toPrint.toString();
    }
  }

  //***************************************************************************//
  //*****************  SYNC CAST **********************************************//
  //***************************************************************************//

  public static void syncCastDomainMessageStarts(int id, boolean normalProcessing)
  {
    if (TRACE)
      print(message(id) + " to be cast " + processingType(normalProcessing));
  }

  public static void syncCastDomainMessageEnded(int id, boolean normalProcessing)
  {
    if (TRACE)
    {
      print(message(id) + " has been casted " +
        (normalProcessing? "and processed":", waiting for tha application to procces it"));
    }
  }

  public static void syncCastDomainMessageProcessed(int id)
  {
    if (TRACE)
      print(message(id) + " has been processed by the application");
  }

  public static void processingDomainMessage(int id)
  {
    if (TRACE)
      print(message(id) + " has been received");
  }

  public static void processedDomainMessage(int id)
  {
    if (TRACE)
      print(message(id) + " has been processed");
  }

  //***************************************************************************//
  //*****************  MESSAGE SEMAPHORE TRACES *******************************//
  //***************************************************************************//

  public static void semaphoreCreated(int id, boolean owned)
  {
    if (TRACE)
      print(semaphore(id) + " created for " + processingType(!owned));
  }

  public static void semaphoreJoining(int id)
  {
    if (TRACE)
      print(semaphore(id) + " to be joined");
  }

  public static void semaphoreJoined(int id)
  {
    if (TRACE)
      print(semaphore(id) + " has been joined");
  }

  public static void semaphoreReleasing(int id)
  {
    if (TRACE)
      print(semaphore(id) + " to be released");
  }

  public static void semaphoreReleased(int id)
  {
    if (TRACE)
      print(semaphore(id) + " has been released");
  }

  public static void semaphoreRemoved(int id)
  {
    if (TRACE)
      print(semaphore(id) + " has been removed");
  }

  public static void semaphoresReleased()
  {
    if (TRACE)
      print("Semaphores are released");
  }


  //***************************************************************************//
  //*****************  REPLICATED MONITOR TRACES ******************************//
  //***************************************************************************//

  public static final String GET_STATE_OPERATION = "GetState";
  public static final String SET_STATE_OPERATION = "SetState";
  public static final String ASSUME_STATE_OPERATION = "AssumeState";
  public static void errorOnStateOperation(String operation)
  {
    if (TRACE_WARNINGS)
    {
      String cause = operation==GET_STATE_OPERATION? "NO" : "ALREADY A";
      print("Error on operation " + operation + " : member has " + cause + " state");
    }
  }

  //***************************************************************************//
  //*****************  GENERAL PURPOSE TRACES *********************************//
  //***************************************************************************//

  public static void print(String message)
  {
    if (logger != null)
      logger.log(message);
  }

  public static void handledException(Exception ex)
  {
    if (TRACE && TRACE_EXCEPTIONS)
    {
      String s;
      if (TRACE_EXCEPTIONS_LONG_FORMAT)
        s = Error.convertExceptionToString(ex);
      else
        s = ex.toString();
      print(s);
    }
  }

  public static void code(String code)
  {
    if (TRACE_CODE)
      print("Code:" + code);
  }

  static String semaphore(int id)
  {
    return "Semaphore {" + id + "} ";
  }

  static String message(int id)
  {
    return "Message {" + id + "} ";
  }

  static String processingType(boolean normalProcessing)
  {
   return " using " + (normalProcessing? "normal processing": "controlled processing");
  }
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
  //+++++++++++++++++  TRACE VARIABLES ++++++++++++++++++++++++++++++++++++++++//
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//

  static final boolean TRACE = sensei.util.Configuration.getSingleton().traceDOMAINS();
  static final boolean TRACE_CODE = sensei.util.Configuration.getSingleton().traceDOMAINSCode();
  static final boolean TRACE_WARNINGS = true;//!!!!!!!!!!!!!
  static final boolean TRACE_EXCEPTIONS = sensei.util.Configuration.getSingleton().traceExceptionsDOMAINS();
  static final boolean TRACE_EXCEPTIONS_LONG_FORMAT = sensei.util.Configuration.getSingleton().traceExceptionsDetailedDOMAINS();

  static final Logger logger = Logger.getLogger(Consts.AREA, TRACE);

}
