/**
  * File: MapStringSetImpl.java
  * Content: Implementation of the MapStringSet interface
  *
  * Author: LuisM Pena
  * Date: 20th November 2001
  * Version: 0.29.00
  * Last change:
  *
  *
  **/

package sensei.gmns;

import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.GroupMember;
import sensei.middleware.gms.Message;
import sensei.middleware.gms.View;
import sensei.middleware.gmns.GMNSnarrower;
import sensei.middleware.gmns.InvalidSetMemberInfo;
import sensei.middleware.gmns.MapStringSetBaseImpl;
import sensei.middleware.gmns.MapStringSetEntry;
import sensei.middleware.gmns.MapStringSetObservable;
import sensei.middleware.gmns.MapStringSetObserver;
import sensei.middleware.gmns.MapStringSetPutMessage;
import sensei.middleware.gmns.MapStringSetRemoveMessage;
import sensei.middleware.gmns.MapStringSetState;
import sensei.middleware.gmns.SetMemberInfo;
import sensei.middleware.domains.DomainGroupHandler;
import sensei.middleware.domains.DynamicSubgroupInfo;
import sensei.middleware.domains.MemberStateException;
import sensei.middleware.domains.State;
import sensei.middleware.domains.StateTransferType;
import sensei.middleware.domains.SubgroupsHandlerException;

import sensei.util.Debug;
import sensei.util.Error;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

class MapStringSetImpl extends MapStringSetBaseImpl
{
//*************************************************************************************//
//**************************** CONSTRUCTOR ********************************************//
//*************************************************************************************//

  private MapStringSetImpl(DomainGroupHandler groupHandler) throws Exception
  {
    Trace.code("MapStringSetImpl.java -> MapStringSetImpl ( DomainGroupHandler groupHandler )");
    this.groupHandler=groupHandler;
    map=new HashMap();
    factory = new GMNSFactory(groupHandler);
    observersHandler = new MapStringSetObserversHandler(thisMapStringSet);
  }

  public int getSubgroupId()
  {
    Trace.code("MapStringSetImpl.java -> int getSubgroupId ( )");
    return subgroupId;
  }

  private void setSubgroupId(int subgroupId)
  {
    Trace.code("MapStringSetImpl.java -> void setSubgroupId ( int subgroupId )");
    this.subgroupId=subgroupId;
    factory.setSubgroupId(subgroupId);
  }

  /**
   * Creates a MapStringSetImpl that must still receive the state.
   * The map can be created after the group has received an event subgroupCreated, or before
   * registering the object in the domain. In the last case, use register after this constructor
   */
  public MapStringSetImpl(int subgroupId, DomainGroupHandler groupHandler) throws Exception
  {
    this(groupHandler);
    Trace.code("MapStringSetImpl.java -> MapStringSetImpl ( int subgroupId , DomainGroupHandler groupHandler )");
    setSubgroupId(subgroupId);
  }

  /**
   * Creates a MapStringSetImpl that registers itself dynamically on the specified groupHandler
   */
  public MapStringSetImpl(DynamicSubgroupInfo dynamicSubgroupInfo, DomainGroupHandler groupHandler)
    throws SubgroupsHandlerException, MemberStateException, Exception
  {
    this(groupHandler);
    Trace.code("MapStringSetImpl.java -> MapStringSetImpl ( DynamicSubgroupInfo dynamicSubgroupInfo , DomainGroupHandler groupHandler )");
    subgroupId = groupHandler.createSubgroup(dynamicSubgroupInfo, thisMapStringSet);
    setSubgroupId(subgroupId);
  }

  /**
   * Register the subgroup statically on the DomainsGroupHandler
   */
  public void register() throws MemberStateException, SubgroupsHandlerException, Exception
  {
    Trace.code("MapStringSetImpl.java -> void register ( )");
    groupHandler.registerSubgroup(subgroupId, thisMapStringSet);
  }

  /**
   * Sends a request to create dynamically a SetMemberInfoImpl subgroup
   */
  public static void register(DynamicSubgroupInfo dynamicSubgroupInfo, DomainGroupHandler groupHandler)
    throws MemberStateException, SubgroupsHandlerException, Exception
  {
    Trace.code("MapStringSetImpl.java -> void register ( DynamicSubgroupInfo dynamicSubgroupInfo , DomainGroupHandler groupHandler )");
    groupHandler.castSubgroupCreation(dynamicSubgroupInfo);
  }

  void freeResources()
  {
    Trace.code("MapStringSetImpl.java -> void freeResources ( )");
    if (map!=null)
    {
      try{deactivate();}catch(Exception ex){}
      map.clear();
      groupHandler=null;
      factory=null;
    }
  }


  //*************************************************************************************
  //**************************** 'PUBLIC' INTERFACE **************************************
  //*************************************************************************************

  public SetMemberInfo put(String group, SetMemberInfo set) throws InvalidSetMemberInfo, MemberStateException
  {
    Trace.code("MapStringSetImpl.java -> SetMemberInfo put ( String group , SetMemberInfo set )");
    SetMemberInfo result=null;
    try
    {
      int subgroupId = groupHandler.getSubgroupId(set);

      if (subgroupId==Consts.INVALID_SUBGROUP)
        throw new InvalidSetMemberInfo();
      factory.castMapPutMessage(group, subgroupId);
      result = doPut(group, set);
      factory.syncMessageCompleted();
    }
    catch(Exception ex){Error.unhandledException(Consts.AREA, ex);}

    return result;
  }

  public SetMemberInfo remove(String group) throws MemberStateException
  {
    Trace.code("MapStringSetImpl.java -> SetMemberInfo remove ( String group )");
    SetMemberInfo result=null;
    try
    {
      factory.castMapRemoveMessage(group);
      result = doRemove(group);
      factory.syncMessageCompleted();
    }
    catch(Exception ex){Error.unhandledException(Consts.AREA, ex);}

    return result;
  }

  public SetMemberInfo get(String group)
  {
    Trace.code("MapStringSetImpl.java -> SetMemberInfo get ( String group )");
    SetMemberInfo ret = null;
    try
    {
      ret = (SetMemberInfo) map.get(group);
    }
    catch(Exception ex){Error.unhandledException(Consts.AREA, ex);}
    return ret;
  }

  public synchronized String[] getKeys()
  {
    Trace.code("MapStringSetImpl.java -> synchronized String [ ] getKeys ( )");
    String ret[] = null;
    try
    {
      int size=map.size();
      ret=new String[size];
      for (Iterator it=map.keySet().iterator();it.hasNext();)
      {
        Debug.assert(size>0,Consts.AREA,"MapStringSetImpl::getKeys");
        ret[--size]=(String) it.next();
      }
    }
    catch(Exception ex){Error.unhandledException(Consts.AREA, ex);}
    return ret;
  }

  //*************************************************************************************
  //**************************** 'PUBLIC' OBSERVER INTERFACE ****************************
  //*************************************************************************************

  public void addObserver(MapStringSetObserver observer)
  {
    Trace.code("MapStringSetImpl.java -> void addObserver ( MapStringSetObserver observer )");
    observersHandler.addObserver(observer);
  }

  public boolean removeObserver(MapStringSetObserver observer)
  {
    Trace.code("MapStringSetImpl.java -> boolean removeObserver ( MapStringSetObserver observer )");
    return observersHandler.removeObserver(observer);
  }

  //*************************************************************************************
  //**************************** 'PRIVATE' STATE INTERFACE ******************************
  //*************************************************************************************

  public State getState()
  {
    Trace.code("MapStringSetImpl.java -> State getState ( )");
    int size = map.size();
    MapStringSetEntry entries[] = new MapStringSetEntry[size];
    if (size>0)
    {
      int i=0;
      Iterator it = map.entrySet().iterator();
      while(it.hasNext())
      {
        Debug.assert(i<size,Consts.AREA,"MapStringSetImpl::GetState::1");
        Map.Entry entry = (Map.Entry) it.next();
        try
        {
          int subgroup=groupHandler.getSubgroupId((SetMemberInfo)entry.getValue());
          Debug.assert(subgroup!=Consts.INVALID_SUBGROUP, Consts.AREA,"MapStringSetImpl::GetState::2");
          entries[i++]=new MapStringSetEntry((String) entry.getKey(),subgroup);
        }
        catch(Exception ex)
        {
          Error.unhandledException(Consts.AREA, ex);
        }
      }
    }
    return factory.createMapState(entries);
  }

  public void setState(State state)
  {
    Trace.code("MapStringSetImpl.java -> void setState ( State state )");
    Debug.assert(state instanceof MapStringSetState, Consts.AREA,"MapStringSetImpl::SetState::1");
    map.clear();

    MapStringSetEntry entries [] = ((MapStringSetState)state).groups;
    int size = entries.length;
    for (int i=0;i<size;i++)
    {
      try
      {
        GroupMember subgroup = groupHandler.getSubgroup(entries[i].value);
        if (subgroup !=null)
        {
          SetMemberInfo set = GMNSnarrower.toSetMemberInfo(subgroup);
          Debug.assert(set!=null, Consts.AREA, "MapStringSetImpl::SetState::2");
          doPut(entries[i].key, set);
        }
      }
      catch(Exception ex)
      {
        Error.unhandledException(Consts.AREA, ex);
      }
    }
  }

  public void assumeState()
  {
    Trace.code("MapStringSetImpl.java -> void assumeState ( )");
    map.clear();
  }

  //*************************************************************************************
  //**************************** 'PRIVATE' GROUP MEMBER INTERFACE ***********************
  //*************************************************************************************

  public void processPTPMessage(int parm1, Message parm2){}
  public void changingView(){}
  public void installView(View parm1){}
  public void memberAccepted(int id, GroupHandler parm2, View parm3)
  {
    Trace.code("MapStringSetImpl.java -> void memberAccepted ( int id , GroupHandler parm2 , View parm3 )");
    memberId=id;
  }
  public void excludedFromGroup()
  {
    Trace.code("MapStringSetImpl.java -> void excludedFromGroup ( )");
    freeResources();
  }
  public void processCastMessage(int sender, Message message)
  {
    Trace.code("MapStringSetImpl.java -> void processCastMessage ( int sender , Message message )");
    try
    {
      if (message instanceof MapStringSetPutMessage)
      {
        MapStringSetPutMessage msg = (MapStringSetPutMessage) message;
        GroupMember subgroup = groupHandler.getSubgroup(msg.value);
        if (subgroup!=null)
        {
          SetMemberInfo set = GMNSnarrower.toSetMemberInfo(subgroup);
          if (set!=null)
            doPut(msg.key, set);
        }
      }
      else if (message instanceof MapStringSetRemoveMessage)
      {
        doRemove(((MapStringSetRemoveMessage) message).key);
      }
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }
  }

  SetMemberInfo doPut(String group, SetMemberInfo set)
  {
    Trace.code("MapStringSetImpl.java -> SetMemberInfo doPut ( String group , SetMemberInfo set )");
    SetMemberInfo ret = null;
    Object obj = null;
    synchronized(this){obj=map.put(group, set);}
    if (obj!=null)
    {
      ret = GMNSnarrower.toSetMemberInfo(obj);
      observersHandler.informRemove(group, ret);
    }
    observersHandler.informPut(group, set);
    return ret;
  }

  SetMemberInfo doRemove(String group)
  {
    Trace.code("MapStringSetImpl.java -> SetMemberInfo doRemove ( String group )");
    SetMemberInfo ret=null;
    Object obj = null;
    synchronized(this){obj=map.remove(group);}
    if (obj!=null)
    {
      ret=GMNSnarrower.toSetMemberInfo(obj);
      observersHandler.informRemove(group, ret);
    }
    return ret;
  }

  synchronized void doClear()
  {
    Trace.code("MapStringSetImpl.java -> synchronized void doClear ( )");
    map.clear();
    observersHandler.informClear();
  }

  //*************************************************************************************
  //**************************** DATA MEMBERS *******************************************
  //*************************************************************************************

  Map map;
  int subgroupId, memberId;
  DomainGroupHandler groupHandler;
  GMNSFactory factory;
  MapStringSetObserversHandler observersHandler;
};
