/**
  * File: SetMemberInfoImpl.java
  * Content: Implementation of the SetMemberInfo interface
  *
  * Author: LuisM Pena
  * Date: 20th November 2001
  * Version: 0.29.00
  * Last change:
  *
  *
  **/

package sensei.gmns;

import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.Message;
import sensei.middleware.gms.View;
import sensei.middleware.gmns.MemberInfo;
import sensei.middleware.gmns.SetMemberInfoBaseImpl;
import sensei.middleware.gmns.SetMemberInfoChangedMessage;
import sensei.middleware.gmns.SetMemberInfoObserver;
import sensei.middleware.gmns.SetMemberInfoState;
import sensei.middleware.domains.DomainGroupHandler;
import sensei.middleware.domains.DynamicSubgroupInfo;
import sensei.middleware.domains.MemberStateException;
import sensei.middleware.domains.State;
import sensei.middleware.domains.StateTransferType;
import sensei.middleware.domains.SubgroupsHandlerException;
import sensei.middleware.util.ObjectsHandling;

import sensei.util.Debug;
import sensei.util.Error;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

class SetMemberInfoImpl extends SetMemberInfoBaseImpl
{
//*************************************************************************************//
//**************************** CONSTRUCTOR ********************************************//
//*************************************************************************************//

  private SetMemberInfoImpl(DomainGroupHandler groupHandler) throws Exception
  {
    Trace.code("SetMemberInfoImpl.java -> SetMemberInfoImpl ( DomainGroupHandler groupHandler )");
    this.groupHandler=groupHandler;
    set=new HashSet();
    factory = new GMNSFactory(groupHandler);
    observersHandler = new SetMemberInfoObserversHandler(thisSetMemberInfo);
  }

  public int getSubgroupId()
  {
    Trace.code("SetMemberInfoImpl.java -> int getSubgroupId ( )");
    return subgroupId;
  }

  private void setSubgroupId(int subgroupId)
  {
    Trace.code("SetMemberInfoImpl.java -> void setSubgroupId ( int subgroupId )");
    this.subgroupId=subgroupId;
    factory.setSubgroupId(subgroupId);
  }

  /**
   * Creates a SetMemberInfoImpl that must still receive the state.
   * The set can be created after the group has received an event subgroupCreated, or before
   * registering the object in the domain. In the last case, use register after this constructor
   */
  public SetMemberInfoImpl(int subgroupId, DomainGroupHandler groupHandler) throws Exception
  {
    this(groupHandler);
    Trace.code("SetMemberInfoImpl.java -> SetMemberInfoImpl ( int subgroupId , DomainGroupHandler groupHandler )");
    setSubgroupId(subgroupId);
  }

  /**
   * Creates a SetMemberInfoImpl that registers itself dynamically on the specified groupHandler
   */
  public SetMemberInfoImpl(DynamicSubgroupInfo dynamicSubgroupInfo, DomainGroupHandler groupHandler)
    throws SubgroupsHandlerException, MemberStateException, Exception
  {
    this(groupHandler);
    Trace.code("SetMemberInfoImpl.java -> SetMemberInfoImpl ( DynamicSubgroupInfo dynamicSubgroupInfo , DomainGroupHandler groupHandler )");
    subgroupId = groupHandler.createSubgroup(dynamicSubgroupInfo, thisSetMemberInfo);
    Debug.assert(subgroupId!=Consts.INVALID_SUBGROUP, Consts.AREA,"SetMemberInfoImpl::SetMemberInfoImpl::1");
    setSubgroupId(subgroupId);
  }

  /**
   * Register the subgroup statically on the DomainsGroupHandler
   */
  public void register() throws MemberStateException, SubgroupsHandlerException, Exception
  {
    Trace.code("SetMemberInfoImpl.java -> void register ( )");
    groupHandler.registerSubgroup(subgroupId, thisSetMemberInfo);
  }

  /**
   * Sends a request to create dynamically a SetMemberInfoImpl subgroup
   */
  public static void register(DynamicSubgroupInfo dynamicSubgroupInfo, DomainGroupHandler groupHandler)
    throws MemberStateException, SubgroupsHandlerException, Exception
  {
    Trace.code("SetMemberInfoImpl.java -> void register ( DynamicSubgroupInfo dynamicSubgroupInfo , DomainGroupHandler groupHandler )");
    groupHandler.castSubgroupCreation(dynamicSubgroupInfo);
  }

  void freeResources()
  {
    Trace.code("SetMemberInfoImpl.java -> void freeResources ( )");
    if (set!=null)
    {
      try{deactivate();}catch(Exception ex){}
      set.clear();
      set=null;
      groupHandler=null;
      factory=null;
    }
  }


  //*************************************************************************************
  //**************************** 'PUBLIC' INTERFACE *************************************
  //*************************************************************************************

  public boolean add(MemberInfo info) throws MemberStateException
  {
    Trace.code("SetMemberInfoImpl.java -> boolean add ( MemberInfo info )");
    boolean result=false;
    try
    {
      factory.castSetMessage(info, true);
      result=addMemberInfo(info);
      factory.syncMessageCompleted();
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }
   return result;
  }

  public synchronized MemberInfo get() throws MemberStateException
  {
    Trace.code("SetMemberInfoImpl.java -> synchronized MemberInfo get ( )");
    Iterator it=set.iterator();
    return it.hasNext()? (MemberInfo) it.next(): null;
  }

  public boolean remove(MemberInfo info) throws MemberStateException
  {
    Trace.code("SetMemberInfoImpl.java -> boolean remove ( MemberInfo info )");
    boolean result=false;
    try
    {
      factory.castSetMessage(info, false);
      result=removeMemberInfo(info);
      factory.syncMessageCompleted();
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }
    return result;
  }

  public synchronized MemberInfo[] toArray()
  {
    Trace.code("SetMemberInfoImpl.java -> synchronized MemberInfo [ ] toArray ( )");
    int size = set.size();
    MemberInfo[] ret = new MemberInfo[size];
    set.toArray(ret);
    return ret;
  }

  //*************************************************************************************
  //**************************** 'PUBLIC' OBSERVER INTERFACE ****************************
  //*************************************************************************************

  public synchronized void addObserver(SetMemberInfoObserver observer)
  {
    Trace.code("SetMemberInfoImpl.java -> synchronized void addObserver ( SetMemberInfoObserver observer )");
    observersHandler.addObserver(observer, getMembers());
  }

  public boolean removeObserver(SetMemberInfoObserver observer)
  {
    Trace.code("SetMemberInfoImpl.java -> boolean removeObserver ( SetMemberInfoObserver observer )");
    return observersHandler.removeObserver(observer);
  }

  //*************************************************************************************
  //**************************** 'PRIVATE' STATE INTERFACE ******************************
  //*************************************************************************************

  public synchronized State getState()
  {
    Trace.code("SetMemberInfoImpl.java -> synchronized State getState ( )");
    return factory.createSetState(getMembers());
  }

  MemberInfo[] getMembers()
  {
    Trace.code("SetMemberInfoImpl.java -> MemberInfo [ ] getMembers ( )");
    int size = set.size();
    MemberInfo members[] = new MemberInfo[size];
    if (size>0)
    {
      int i=0;
      Iterator it = set.iterator();
      while(it.hasNext())
      {
        Debug.assert(i<size,Consts.AREA,"SetMemberInfoImpl::GetMembers");
        members[i++]=(MemberInfo) it.next();
      }
    }
    return members;
  }

  public synchronized void setState(State state)
  {
    Trace.code("SetMemberInfoImpl.java -> synchronized void setState ( State state )");
    Debug.assert(state instanceof SetMemberInfoState, Consts.AREA,"SetMemberInfoImpl::SetState");
    set.clear();
    MemberInfo members[]=((SetMemberInfoState)state).members;
    int size = members.length;
    for (int i=0;i<size;i++)
      addMemberInfo(members[i]);
  }

  public synchronized void assumeState()
  {
    Trace.code("SetMemberInfoImpl.java -> synchronized void assumeState ( )");
    set.clear();
  }
  //*************************************************************************************
  //**************************** 'PRIVATE' GROUP MEMBER INTERFACE ***********************
  //*************************************************************************************

  public void processPTPMessage(int parm1, Message parm2){}
  public void changingView(){}
  public void installView(View parm1){}
  public void memberAccepted(int id, GroupHandler parm2, View parm3)
  {
    Trace.code("SetMemberInfoImpl.java -> void memberAccepted ( int id , GroupHandler parm2 , View parm3 )");
    memberId=id;
  }
  public void excludedFromGroup()
  {
    Trace.code("SetMemberInfoImpl.java -> void excludedFromGroup ( )");
    freeResources();
  }
  public void processCastMessage(int sender, Message message)
  {
    Trace.code("SetMemberInfoImpl.java -> void processCastMessage ( int sender , Message message )");
    if (message instanceof SetMemberInfoChangedMessage)
    {
      synchronized(this)
      {
        SetMemberInfoChangedMessage msg = (SetMemberInfoChangedMessage) message;
        if (msg.add)
          addMemberInfo(msg.info);
        else
          removeMemberInfo(msg.info);
      }
    }
  }

  synchronized boolean addMemberInfo(MemberInfo memberInfo)
  {
    Trace.code("SetMemberInfoImpl.java -> synchronized boolean addMemberInfo ( MemberInfo memberInfo )");
    boolean ret = set.add(memberInfo);
    if (ret)
      observersHandler.informPut(memberInfo);
    return ret;
  }

  synchronized boolean removeMemberInfo(MemberInfo memberInfo)
  {
    Trace.code("SetMemberInfoImpl.java -> synchronized boolean removeMemberInfo ( MemberInfo memberInfo )");
    Iterator it = set.iterator();
    while(it.hasNext())
    {
      MemberInfo content = (MemberInfo) it.next();
      if (ObjectsHandling.areEquivalent(memberInfo.handler, content.handler)) //it is only checked the handler
      {
        it.remove();
        observersHandler.informRemove(memberInfo);
        return true;
      }
    }
    return false;
  }

  //*************************************************************************************
  //**************************** DATA MEMBERS *******************************************
  //*************************************************************************************

  Set set; //MemberInfo!
  int subgroupId, memberId;
  DomainGroupHandler groupHandler;
  GMNSFactory factory;
  SetMemberInfoObserversHandler observersHandler;
};
