/**
  * File: AllowTokenRecoveryChannelRequester.java
  * Content: Defines a channel requester to request the token recovery
  * Author: LuisM Pena
  * Date: 1st August 2000
  * Version: 0.97.00
  * Last change:
  *
  **/

package sensei.gms;

import sensei.middleware.gms.AllowTokenRecoveryAnswer;
import sensei.middleware.gms.InternalViewId;
import sensei.middleware.gms.MessageId;
import sensei.middleware.gms.SenseiGMSMember;

import sensei.util.Debug;

//*************************************************************************************//
//**************************** ALLOW TOKEN RECOVERY CHANNEL REQUESTER *****************//
//*************************************************************************************//

/**
  * Channel Requester used to send requests to try to regenerate the token
  **/
class AllowTokenRecoveryChannelRequester implements ChannelRequester
{

  public AllowTokenRecoveryChannelRequester(int thisMember, InternalViewId viewId, MessageId messageId)
  {
    Trace.code("AllowTokenRecoveryChannelRequester.java -> AllowTokenRecoveryChannelRequester ( int thisMember , InternalViewId viewId , MessageId messageId )");

    Debug.assert(messageId!=null && viewId!=null, Consts.GMS, "AllowTokenRecoveryChannelRequester:AllowTokenRecoveryChannelRequester:1");

    this.thisMember 			= thisMember;
    this.viewId 		   		= viewId;
    this.messageId     		= messageId;
    expectingFirstMessage	= true;
    answer             		= true;
    processLastMessage 		= false;
  }

  public boolean invokeMember(CommunicationChannel channel)
  {
    Trace.code("AllowTokenRecoveryChannelRequester.java -> boolean invokeMember ( CommunicationChannel channel )");
    Debug.assert(channel!=null, Consts.GMS,"AllowTokenRecoveryChannelRequester:invokeMember:1");

    boolean ret;
    SenseiGMSMember channelMember = channel.getLink();

    Debug.assert(channelMember!=null,Consts.GMS,"AllowTokenRecoveryChannelRequester:invokeMember:2");

    try
    {
      AllowTokenRecoveryAnswer retAnswer = channelMember.allowTokenRecovery(thisMember, viewId);
      ret = retAnswer.validCommunication  &&  channel.isReliable();
      if (ret)
      {
        answer = answer && retAnswer.recoveryAllowed;
        if (answer && !processLastMessage)
        {
          boolean copy = false;
          if (expectingFirstMessage)
          {
            copy 									= true;
            expectingFirstMessage = false;
          }
          else
          {
            //if the message returned is different of 'messageId' and it's not null,
            //the lasMessageMust been processed
            processLastMessage =
                    ((retAnswer.lastMessage.view != messageId.view) || (retAnswer.lastMessage.id != messageId.id))
                && 	(retAnswer.lastMessage.view != 0);

            //if the message returned is bigger, it must be last message to be processed
            //note that it's not checked if there are differences on the messageIds (for
            //example, two messages distanced by more than two steps, this will be recogniced
            //later on the recovery
            if (processLastMessage)
              if ((retAnswer.lastMessage.view > messageId.view)
                || ((retAnswer.lastMessage.view == messageId.view) && (retAnswer.lastMessage.id > messageId.id)))
                copy = true;
          }
          if (copy)
          {
            messageId.view = retAnswer.lastMessage.view;
            messageId.id   = retAnswer.lastMessage.id;
          }
        }
      }
    }
    catch (Exception ex)
    {
      ret=false;
      Trace.handledException(ex);
    }
    return ret;
  }

  /**
    * Returns the answer given by the channels. If it's true, the token can be regenerated,
    * and in that case processLastMessage will be set to true if the last message in the
    * group has still to be processed (confirmation)
    * It only takes into account those answers which members have replied yes to the
    * communication (they can still answer not as answer)
    **/
  public boolean getAnswer(OutBoolean processLastMessage)
  {
    Trace.code("AllowTokenRecoveryChannelRequester.java -> boolean getAnswer ( OutBoolean processLastMessage )");
    if (answer)
      processLastMessage.value = this.processLastMessage;
    return answer;
  }

  int thisMember;
  InternalViewId viewId;
  boolean answer;	//token can be recovered or not
  boolean processLastMessage;	//true if not every member has processed the same messages
  boolean expectingFirstMessage;
  MessageId messageId;
}
