/**
  * File: RestartableThread.java
  * Content: Class allowing a 'restart' method-work
  * Author: LuisM Pena
  * Date: 8nd July 2000
  * Version: 0.97.00
  * Last change:
  *   15th February 2002: changed the synchronization, it was wrong
  *
  * Comments: the synchronized code is done with a Lock, an external
  *			object, to avoid conflicts with synchronized code in the classes
  *			inheriting from RestartableThread. A much better approach would be
  *			to implement RestartableThread to be used and not inherited.
  *
  **/

package sensei.util;

/**
  * A thread can not be started more than once.
  * This class introduces kind of that functionality, using activate instead
  * of start. This thread is always started, and it must be stopped explicitely
  * with the end method
  **/
public class RestartableThread extends Thread implements Rerunnable
{
  public RestartableThread (ThreadGroup group, String name)
  {
    super(group,name);
    constructor(null);
  }

  public RestartableThread()
  {
    constructor(null);
  }

  public RestartableThread(Rerunnable target)
  {
    constructor(target);
  }

  /**
    * Called when the thread is finally activated.
    * @param taskParam the argument passed to activate
    * @returns true if the thread must continue its execution, or
    * false if it must be ended
    **/
  public boolean activated(Object taskParam)
  {
    return (target==null)? false :target.activated(taskParam);
  }

  /**
    * Checks wether the thread is activated or not
    * @returns true if the thread is activated
    **/
  public boolean isActivated()
  {
    return started;
  }

  /**
    * Activates the thread, blocking only the number os seconds if it's already started.
    * If the thread can not be activated in the time given, the method returns false
    * @param seconds number of seconds to wait
    * @param taskParam the argument passed to activate
    * @exception InterruptedException
    * @returns true if the process has been activated
    **/
  public boolean activate(int seconds, Object taskParam) throws InterruptedException
  {
    boolean ret = false;
    synchronized(lock)
    {
      if (!stopped)
      {
        if (started)
          lock.wait(seconds);

        if (!started)
        {
          started=true;
          this.taskParam = taskParam;
          lock.notifyAll();
          ret=true;
        }
      }
    }

    return ret;
  }

  /**
    * Ends the thread. Note that it is not stopped, if it is also required to
    * stop it, the caller should use after end() a call to join(), as some
    * time will last until the thread is effectively ended.
    **/
  public void end()
  {
    synchronized(lock)
    {
      if (!stopped)
      {
        stopped = true;
        lock.notifyAll();
      }
    }
  }

  /**
    * Checks if the thread is still valid and can be activated
    * @returns true if the thread has finished and cannot be reactivated
    **/
  public boolean isEnded()
  {
    return stopped;
  }

  /**
    * Blocks until the current process finishes
    * @param seconds number of seconds to wait
    * @exception InterruptedException
    * @returns true if the process has been awaited (not time last)
    **/
  public boolean waitCurrent(int seconds) throws InterruptedException
  {
    boolean ret;
    synchronized(lock)
    {
      if (!stopped && started)
        lock.wait(seconds);
      ret=!started && !stopped;
    }
    return ret;
  }


  public final void run()
  {
    try
    {
      while (!stopped)
      {
        synchronized(lock)
        {
          while (!(started || stopped))
            lock.wait();
        }
        if (!stopped)
          if (!activated(taskParam))
            stopped=true;

        synchronized(lock)
        {
          started   = false;
          taskParam = null;
          lock.notifyAll();
        }

      }
    }
    catch(Exception ex)
    {
      exception=stopped=true;
      started=false;
      Error.unhandledException(Consts.AREA_NAME, ex);
    }
    taskParam = null;
  }

  void constructor(Rerunnable target)
  {
    this.target = target;
    started			=	false;
    stopped			=	false;
    exception		=	false;
    taskParam		=	null;
    lock      	= new Boolean(true);	//any object
    start();
  }

  boolean 		started;
  boolean 		stopped;
  boolean 		exception;

  Object 			taskParam;
  Object  		lock;
  Rerunnable	target;
}
